import datetime
import unittest
from unittest import mock

from lgrez import config, bdd
from lgrez.features import open_close
from test import mock_discord, mock_bdd, mock_env


class TestOpenCloseFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.open_close utility functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    @unittest.SkipTest
    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.gestion_actions.get_actions")
    async def test_recup_joueurs(self, getact_patch):
        """Unit tests for open_close.recup_joueurs function."""
        # async def recup_joueurs(quoi, qui, heure=None)
        recup_joueurs = open_close.recup_joueurs
        mock_bdd.add_campsroles()
        i = 0
        joueurs = []
        for vote_condamne_ in [None, "non défini", "oh"]:
            for vote_maire_ in [None, "non défini", "ah"]:
                for vote_loups_ in [None, "non défini", "eh"]:
                    for votant_village in [True, False]:
                        for votant_loups in [True, False]:
                            for role_actif in [True, False]:
                                j = bdd.Joueur(
                                    discord_id=i, chan_id_=i,
                                    nom=f"Joueur{i}",
                                    vote_condamne_=vote_condamne_,
                                    vote_maire_=vote_maire_,
                                    vote_loups_=vote_loups_,
                                    votant_village=votant_village,
                                    votant_loups=votant_loups,
                                    role_actif=role_actif
                                )
                                joueurs.append(j)
                                i += 1
        bdd.Joueur.add(*joueurs)

        # bad quoi
        with self.assertRaises(ValueError):
            await recup_joueurs("bloup", "cond")
        getact_patch.assert_not_called()

        # bad qui
        with self.assertRaises(ValueError):
            await recup_joueurs("open", "bzzt")
        getact_patch.assert_not_called()
        with self.assertRaises(ValueError):
            await recup_joueurs("open", "19")     # non-existing action
        getact_patch.assert_not_called()

        # qui = "cond"
        results = await recup_joueurs("open", "cond")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.votant_village and joueur.vote_condamne_ is None:
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        results = await recup_joueurs("remind", "cond")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.vote_condamne_ == "non défini":
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        results = await recup_joueurs("close", "cond")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.vote_condamne_ is not None:
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        # qui = "maire"
        results = await recup_joueurs("open", "maire")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.votant_village and joueur.vote_maire_ is None:
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        results = await recup_joueurs("remind", "maire")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.vote_maire_ == "non défini":
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        results = await recup_joueurs("close", "maire")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.vote_maire_ is not None:
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        # qui = "loups"
        results = await recup_joueurs("open", "loups")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.votant_loups and joueur.vote_loups_ is None:
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        results = await recup_joueurs("remind", "loups")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.vote_loups_ == "non défini":
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        results = await recup_joueurs("close", "loups")
        getact_patch.assert_not_called()
        for joueur in joueurs:
            if joueur.vote_loups_ is not None:
                self.assertIn(joueur, results)
            else:
                self.assertNotIn(joueur, results)

        # qui = "action"
        with self.assertRaises(ValueError):
            await recup_joueurs("open", "action")
        getact_patch.assert_not_called()

        with self.assertRaises(ValueError):
            await recup_joueurs("open", "action", 15)
        getact_patch.assert_not_called()

        bdd.BaseAction(slug="ouiZ").add()
        bdd.BaseAction(slug="nonZ").add()
        bdd.BaseAction(slug="lalaZ").add()
        action1 = bdd.Action(id=1, joueur=joueurs[0], _base_slug="ouiZ")
        action2 = bdd.Action(id=23, joueur=joueurs[0], _base_slug="nonZ")
        action3 = bdd.Action(id=72, joueur=joueurs[1], _base_slug="lalaZ")
        bdd.Action.add(action1, action2, action3)

        getact_patch.return_value = [action1, action2, action3]
        results = await recup_joueurs("open", "action", "15h23")
        self.assertEqual(results, {joueurs[0]: [action1, action2],
                                   joueurs[1]: [action3]})
        getact_patch.assert_called_once_with(
            "open", bdd.ActionTrigger.temporel, datetime.time(15, 23)
        )
        getact_patch.reset_mock()

        results = await recup_joueurs("close", "action", "7h")
        self.assertEqual(results, {joueurs[0]: [action1, action2],
                                   joueurs[1]: [action3]})
        getact_patch.assert_called_once_with(
            "close", bdd.ActionTrigger.temporel, datetime.time(7, 0)
        )
        getact_patch.reset_mock()

        results = await recup_joueurs("remind", "action", "23h12")
        self.assertEqual(results, {joueurs[0]: [action1, action2],
                                   joueurs[1]: [action3]})
        getact_patch.assert_called_once_with(
            "remind", bdd.ActionTrigger.temporel, datetime.time(23, 12)
        )
        getact_patch.reset_mock()

        # qui = id
        with self.assertRaises(ValueError):
            await recup_joueurs("open", "3")    # Non-existing action
        getact_patch.assert_not_called()

        action3.decision_ = None
        action3.base.trigger_debut = bdd.ActionTrigger.perma
        action3.update()
        results = await recup_joueurs("open", "72")
        self.assertEqual(results, {joueurs[1]: [action3]})
        getact_patch.assert_not_called()

        action3.decision_ = "blabla"
        action3.update()
        results = await recup_joueurs("open", "72")
        self.assertEqual(results, {joueurs[1]: [action3]})
        getact_patch.assert_not_called()

        action3.decision_ = None
        action3.base.trigger_debut = bdd.ActionTrigger.mot_mjs
        action3.update()
        results = await recup_joueurs("open", "72")
        self.assertEqual(results, {joueurs[1]: [action3]})
        getact_patch.assert_not_called()

        action3.decision_ = "blabla"
        action3.update()
        results = await recup_joueurs("open", "72")
        self.assertEqual(results, {})
        getact_patch.assert_not_called()

        results = await recup_joueurs("remind", "72")
        self.assertEqual(results, {})
        getact_patch.assert_not_called()

        action3.decision_ = "rien"
        action3.update()
        results = await recup_joueurs("remind", "72")
        self.assertEqual(results, {joueurs[1]: [action3]})
        getact_patch.assert_not_called()

        results = await recup_joueurs("close", "72")
        self.assertEqual(results, {joueurs[1]: [action3]})
        getact_patch.assert_not_called()

        action3.decision_ = "blabla"
        action3.update()
        results = await recup_joueurs("close", "72")
        self.assertEqual(results, {joueurs[1]: [action3]})
        getact_patch.assert_not_called()

        action3.decision_ = None
        action3.update()
        results = await recup_joueurs("close", "72")
        self.assertEqual(results, {})
        getact_patch.assert_not_called()


    @mock_bdd.patch_db      # Empty database for this method
    async def test__do_refill(self):
        """Unit tests for open_close._do_refill function."""
        # async def _do_refill(motif, actions)
        _do_refill = open_close._do_refill

        config.refills_full = ["uikz"]
        config.refills_one = ["forgebonk", "rebootax", "divvvvvvin"]
        config.refills_divins = ["divvvvvvin"]

        mock_bdd.add_campsroles()
        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        joueur2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2")
        joueurs = [joueur1, joueur2]
        bdd.Joueur.add(*joueurs)
        actions = {}
        i = 0
        for kw in [*config.refills_full, *config.refills_one]:
            base = f"baz_{kw}"
            bdd.BaseAction(slug=base, refill=kw, base_charges=5 + i).add()
            i += 1
            actions[kw] = [
                bdd.Action(joueur=joueur1, _base_slug=base, charges=3),
                bdd.Action(joueur=joueur2, _base_slug=base, charges=3),
            ]
        action_none = bdd.Action(joueur=joueur1, _base_slug=base, charges=None)
        all_actions = {ac for acs in actions.values() for ac in acs}
        bdd.Action.add(*all_actions, action_none)

        # motif = full
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await _do_refill("uikz", list(all_actions))
        for action in all_actions:
            chan = chan1 if action.joueur == joueur1 else chan2
            sent = "\n".join(call.args[0] for call in chan.send.call_args_list)
            self.assertEqual(action.charges, action.base.base_charges)
            self.assertIn(str(action.base.slug), sent)
            self.assertIn(str(action.base.base_charges), sent)

        # motif = not full
        for action in all_actions:
            action.charges = 3
        config.session.commit()

        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await _do_refill("forgebonk", list(all_actions))
        for action in all_actions:
            chan = chan1 if action.joueur == joueur1 else chan2
            sent = "\n".join(call.args[0] for call in chan.send.call_args_list)
            self.assertEqual(action.charges, 4)
            self.assertIn(str(action.base.slug), sent)
            self.assertIn("4", sent)

        # motif = full, already full action
        bdd.BaseAction(slug="oui", base_charges=7).add()
        fa = bdd.Action(joueur=joueur1, _base_slug="oui", charges=7)
        fa.add()
        with mock_discord.mock_members_and_chans(joueur1):
            chan1 = joueur1.private_chan
            await _do_refill("uikz", [fa])
        self.assertEqual(fa.charges, 7)
        chan1.send.assert_not_called()

        # motif = not full, already full action
        with mock_discord.mock_members_and_chans(joueur1):
            chan1 = joueur1.private_chan
            await _do_refill("forgebonk", [fa])
        self.assertEqual(fa.charges, 8)
        chan1.send.assert_called_once()

        # perma action with no charges
        bdd.BaseAction(slug="non", trigger_debut=bdd.ActionTrigger.perma,
                       base_charges=7).add()
        pa = bdd.Action(id=13, joueur=joueur1, _base_slug="non", charges=0)
        pa.add()
        with mock_discord.mock_members_and_chans(joueur1):
            await _do_refill("forgebonk", [pa])
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open 13")
        self.assertEqual(taches[0].action, pa)



@unittest.SkipTest
class TestOpenClose(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.open_close commands."""

    def setUp(self):
        mock_discord.mock_config()
        self.cog = open_close.OpenClose(config.bot)

    def tearDown(self):
        mock_discord.unmock_config()

    @mock_bdd.patch_db      # Empty database for this method
    @mock_discord.interact()
    @mock.patch("lgrez.features.open_close.recup_joueurs")  # tested before
    @mock.patch("lgrez.features.gestion_actions.open_action")
    async def test_open(self, oa_patch, rj_patch):
        """Unit tests for !open command."""
        # async def open(self, ctx, qui, heure=None, heure_chain=None)
        open = self.cog.open
        mock_bdd.add_campsroles()
        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                             votant_village=True, votant_loups=True)
        joueur2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                             votant_village=True, votant_loups=True)
        joueur3 = bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3")
        joueurs = [joueur1, joueur2, joueur3]
        bdd.Joueur.add(*joueurs)
        haros = [bdd.CandidHaro(joueur=joueur2, type=bdd.CandidHaroType.haro),
                 bdd.CandidHaro(joueur=joueur3, type=bdd.CandidHaroType.haro)]
        candids = [bdd.CandidHaro(joueur=joueur1,
                                  type=bdd.CandidHaroType.candidature),
                   bdd.CandidHaro(joueur=joueur3,
                                  type=bdd.CandidHaroType.candidature)]
        bdd.CandidHaro.add(*haros, *candids)
        bases = [bdd.BaseAction(slug=trigger.name, trigger_debut=trigger)
                 for trigger in bdd.ActionTrigger]
        ac_oc = [bdd.Action(joueur=joueur1, base=base)
                 for joueur in joueurs for base in bases]


        # bad qui
        ctx = mock_discord.get_ctx(open, "bzeepzopsl")
        rj_patch.side_effect = ValueError
        with self.assertRaises(ValueError):
            await ctx.invoke()
        rj_patch.assert_called_once_with("open", "bzeepzopsl", None)
        rj_patch.reset_mock(side_effect=True)
        oa_patch.assert_not_called()

        # qui = "cond", heure = None, heure_chain = None
        ctx = mock_discord.get_ctx(open, "cond")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("open", "cond", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Ouverture vote
        self.assertEqual(joueur1.vote_condamne_, "non défini")
        self.assertEqual(joueur2.vote_condamne_, "non défini")
        self.assertIsNone(joueur3.vote_condamne_)
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("vote pour le condamné du jour est ouvert",
                      chan1.send.call_args.args[0])
        (await chan1.send()).add_reaction.assert_called_once_with(
            config.Emoji.bucher)
        chan2.send.assert_called_once()
        self.assertIn("vote pour le condamné du jour est ouvert",
                      chan2.send.call_args.args[0])
        (await chan2.send()).add_reaction.assert_called_once_with(
            config.Emoji.bucher)
        # Ouverture open_cond actions
        self.assertEqual(oa_patch.call_count, 3)
        self.assertEqual(
            [ac for ac in ac_oc
             if ac.base.trigger_debut == bdd.ActionTrigger.open_cond],
            [call.args[0] for call in oa_patch.call_args_list])
        oa_patch.reset_mock()
        # Réinitialisation haros
        self.assertFalse(bdd.CandidHaro.query.filter_by(
            type=bdd.CandidHaroType.haro
        ).all())
        self.assertEqual(bdd.CandidHaro.query.filter_by(
            type=bdd.CandidHaroType.candidature
        ).all(), candids)
        config.Channel.haros.send.assert_called_once
        self.assertIn("Nouveau vote",
                      config.Channel.haros.send.call_args.args[0])
        config.Channel.haros.send.reset_mock()
        # heure / heure_chain
        self.assertFalse(bdd.Tache.query.all())

        # qui = "cond", heure = 15h12, heure_chain = None
        ctx = mock_discord.get_ctx(open, "cond", heure="15h12")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        oa_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual({tache.commande for tache in taches},
                         {"!remind cond", "!close cond"})
        bdd.Tache.delete(*taches)

        # qui = "cond", heure = 15h12, heure_chain = 7h
        ctx = mock_discord.get_ctx(open, "cond", heure="15h12",
                                   heure_chain="7h")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        oa_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual({tache.commande for tache in taches},
                         {"!remind cond", "!close cond 7h 15h12"})
        bdd.Tache.delete(*taches)

        # ---- maire ----
        bdd.CandidHaro.delete(*haros, *candids)
        haros = [bdd.CandidHaro(joueur=joueur2, type=bdd.CandidHaroType.haro),
                 bdd.CandidHaro(joueur=joueur3, type=bdd.CandidHaroType.haro)]
        candids = [bdd.CandidHaro(joueur=joueur1,
                                  type=bdd.CandidHaroType.candidature),
                   bdd.CandidHaro(joueur=joueur3,
                                  type=bdd.CandidHaroType.candidature)]
        bdd.CandidHaro.add(*haros, *candids)

        # qui = "maire", heure = None, heure_chain = None
        ctx = mock_discord.get_ctx(open, "maire")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("open", "maire", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Ouverture vote
        self.assertEqual(joueur1.vote_maire_, "non défini")
        self.assertEqual(joueur2.vote_maire_, "non défini")
        self.assertIsNone(joueur3.vote_maire_)
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("vote pour l'élection du maire est ouvert",
                      chan1.send.call_args.args[0])
        (await chan1.send()).add_reaction.assert_called_once_with(
            config.Emoji.maire)
        chan2.send.assert_called_once()
        self.assertIn("vote pour l'élection du maire est ouvert",
                      chan2.send.call_args.args[0])
        (await chan2.send()).add_reaction.assert_called_once_with(
            config.Emoji.maire)
        # Ouverture open_maire actions
        self.assertEqual(oa_patch.call_count, 3)
        self.assertEqual(
            [ac for ac in ac_oc
             if ac.base.trigger_debut == bdd.ActionTrigger.open_maire],
            [call.args[0] for call in oa_patch.call_args_list])
        oa_patch.reset_mock()
        # Réinitialisation haros
        self.assertFalse(bdd.CandidHaro.query.filter_by(
            type=bdd.CandidHaroType.candidature
        ).all())
        self.assertEqual(bdd.CandidHaro.query.filter_by(
            type=bdd.CandidHaroType.haro
        ).all(), haros)
        config.Channel.haros.send.assert_called_once
        self.assertIn("Nouveau vote",
                      config.Channel.haros.send.call_args.args[0])
        config.Channel.haros.send.reset_mock()
        # heure / heure_chain
        self.assertFalse(bdd.Tache.query.all())

        # qui = "maire", heure = 15h12, heure_chain = None
        ctx = mock_discord.get_ctx(open, "maire", heure="15h12")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        oa_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual({tache.commande for tache in taches},
                         {"!remind maire", "!close maire"})
        bdd.Tache.delete(*taches)

        # qui = "maire", heure = 15h12, heure_chain = 7h
        ctx = mock_discord.get_ctx(open, "maire", heure="15h12",
                                   heure_chain="7h")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        oa_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual({tache.commande for tache in taches},
                         {"!remind maire", "!close maire 7h 15h12"})
        bdd.Tache.delete(*taches)

        # ---- loups ----
        bdd.CandidHaro.delete(*haros, *candids)
        haros = [bdd.CandidHaro(joueur=joueur2, type=bdd.CandidHaroType.haro),
                 bdd.CandidHaro(joueur=joueur3, type=bdd.CandidHaroType.haro)]
        candids = [bdd.CandidHaro(joueur=joueur1,
                                  type=bdd.CandidHaroType.candidature),
                   bdd.CandidHaro(joueur=joueur3,
                                  type=bdd.CandidHaroType.candidature)]
        bdd.CandidHaro.add(*haros, *candids)

        # qui = "loups", heure = None, heure_chain = None
        ctx = mock_discord.get_ctx(open, "loups")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("open", "loups", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Ouverture vote
        self.assertEqual(joueur1.vote_loups_, "non défini")
        self.assertEqual(joueur2.vote_loups_, "non défini")
        self.assertIsNone(joueur3.vote_loups_)
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("vote pour la victime de cette nuit est ouvert",
                      chan1.send.call_args.args[0])
        (await chan1.send()).add_reaction.assert_called_once_with(
            config.Emoji.lune)
        chan2.send.assert_called_once()
        self.assertIn("vote pour la victime de cette nuit est ouvert",
                      chan2.send.call_args.args[0])
        (await chan2.send()).add_reaction.assert_called_once_with(
            config.Emoji.lune)
        # Ouverture open_loups actions
        self.assertEqual(oa_patch.call_count, 3)
        self.assertEqual(
            [ac for ac in ac_oc
             if ac.base.trigger_debut == bdd.ActionTrigger.open_loups],
            [call.args[0] for call in oa_patch.call_args_list])
        oa_patch.reset_mock()
        # Non-réinitialisation haros
        self.assertEqual(len(bdd.CandidHaro.query.all()), 4)
        config.Channel.haros.send.assert_not_called()
        # heure / heure_chain
        self.assertFalse(bdd.Tache.query.all())

        # qui = "loups", heure = 15h12, heure_chain = None
        ctx = mock_discord.get_ctx(open, "loups", heure="15h12")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        oa_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual({tache.commande for tache in taches},
                         {"!remind loups", "!close loups"})
        bdd.Tache.delete(*taches)

        # qui = "loups", heure = 15h12, heure_chain = 7h
        ctx = mock_discord.get_ctx(open, "loups", heure="15h12",
                                   heure_chain="7h")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        oa_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual({tache.commande for tache in taches},
                         {"!remind loups", "!close loups 7h 15h12"})

        # ---- action ----
        bdd.BaseAction(slug="ouiz", trigger_debut=bdd.ActionTrigger.temporel,
                       heure_debut=datetime.time(15, 12)).add()
        action1 = bdd.Action(_base_slug="ouiz", joueur=joueur1)
        action2 = bdd.Action(_base_slug="ouiz", joueur=joueur1)
        action3 = bdd.Action(_base_slug="ouiz", joueur=joueur2)
        bdd.Action.add(action1, action2, action3)

        # qui = "action", heure = "15h12"
        ctx = mock_discord.get_ctx(open, "action", "15h12")
        rj_patch.return_value = {joueur1: [action1, action2],
                                 joueur2: [action3]}
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("open", "action", "15h12")
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Ouverture actions
        self.assertEqual(oa_patch.call_count, 3)
        self.assertEqual(
            [action1, action2, action3],
            [call.args[0] for call in oa_patch.call_args_list])
        oa_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock_discord.interact()
    @mock.patch("lgrez.features.open_close.recup_joueurs")  # tested before
    @mock.patch("lgrez.features.gestion_actions.close_action")
    async def test_close(self, ca_patch, rj_patch):
        """Unit tests for !close command."""
        # async def close(self, ctx, qui, heure=None, heure_chain=None)
        close = self.cog.close
        mock_bdd.add_campsroles()
        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                             votant_village=True, votant_loups=True,
                             vote_condamne_="zeret", vote_maire_="goo",
                             vote_loups_="kowwwia")
        joueur2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                             votant_village=True, votant_loups=True,
                             vote_condamne_="zeret", vote_maire_="goo",
                             vote_loups_="kowwwia")
        joueur3 = bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3",
                             vote_condamne_="zeret", vote_maire_="goo",
                             vote_loups_="kowwwia")
        joueurs = [joueur1, joueur2, joueur3]
        bdd.Joueur.add(*joueurs)
        bases = [bdd.BaseAction(slug=trigger.name, trigger_debut=trigger)
                 for trigger in bdd.ActionTrigger]
        ac_oc = [bdd.Action(joueur=joueur1, base=base)
                 for joueur in joueurs for base in bases]

        # bad qui
        ctx = mock_discord.get_ctx(close, "bzeepzopsl")
        rj_patch.side_effect = ValueError
        with self.assertRaises(ValueError):
            await ctx.invoke()
        rj_patch.assert_called_once_with("close", "bzeepzopsl", None)
        rj_patch.reset_mock(side_effect=True)
        ca_patch.assert_not_called()

        # qui = "cond", heure = None, heure_chain = None
        ctx = mock_discord.get_ctx(close, "cond")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("close", "cond", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Fermeture vote
        self.assertIsNone(joueur1.vote_condamne_)
        self.assertIsNone(joueur2.vote_condamne_)
        self.assertEqual(joueur3.vote_condamne_, "zeret")
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("Fin du vote pour le condamné",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("Fin du vote pour le condamné",
                      chan2.send.call_args.args[0])
        # Fermeture close_cond actions
        self.assertEqual(ca_patch.call_count, 3)
        self.assertEqual(
            [ac for ac in ac_oc
             if ac.base.trigger_debut == bdd.ActionTrigger.close_cond],
            [call.args[0] for call in ca_patch.call_args_list])
        ca_patch.reset_mock()
        # heure / heure_chain
        self.assertFalse(bdd.Tache.query.all())

        # qui = "cond", heure = 15h12, heure_chain = None
        ctx = mock_discord.get_ctx(close, "cond", heure="15h12")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        ca_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open cond")
        bdd.Tache.delete(*taches)

        # qui = "cond", heure = 15h12, heure_chain = 7h
        ctx = mock_discord.get_ctx(close, "cond", heure="15h12",
                                   heure_chain="7h")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        ca_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open cond 7h 15h12")
        bdd.Tache.delete(*taches)

        # ---- maire ----
        # qui = "maire", heure = None, heure_chain = None
        ctx = mock_discord.get_ctx(close, "maire")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("close", "maire", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Fermeture vote
        self.assertIsNone(joueur1.vote_maire_)
        self.assertIsNone(joueur2.vote_maire_)
        self.assertEqual(joueur3.vote_maire_, "goo")
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("Fin du vote pour le maire",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("Fin du vote pour le maire",
                      chan2.send.call_args.args[0])
        # Fermeture close_maire actions
        self.assertEqual(ca_patch.call_count, 3)
        self.assertEqual(
            [ac for ac in ac_oc
             if ac.base.trigger_debut == bdd.ActionTrigger.close_maire],
            [call.args[0] for call in ca_patch.call_args_list])
        ca_patch.reset_mock()
        # heure / heure_chain
        self.assertFalse(bdd.Tache.query.all())

        # qui = "maire", heure = 15h12, heure_chain = None
        ctx = mock_discord.get_ctx(close, "maire", heure="15h12")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        ca_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open maire")
        bdd.Tache.delete(*taches)

        # qui = "maire", heure = 15h12, heure_chain = 7h
        ctx = mock_discord.get_ctx(close, "maire", heure="15h12",
                                   heure_chain="7h")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        ca_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open maire 7h 15h12")
        bdd.Tache.delete(*taches)

        # ---- loups ----
        # qui = "loups", heure = None, heure_chain = None
        ctx = mock_discord.get_ctx(close, "loups")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("close", "loups", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Fermeture vote
        self.assertIsNone(joueur1.vote_loups_)
        self.assertIsNone(joueur2.vote_loups_)
        self.assertEqual(joueur3.vote_loups_, "kowwwia")
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("Fin du vote pour la victime",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("Fin du vote pour la victime",
                      chan2.send.call_args.args[0])
        # Fermeture close_loups actions
        self.assertEqual(ca_patch.call_count, 3)
        self.assertEqual(
            [ac for ac in ac_oc
             if ac.base.trigger_debut == bdd.ActionTrigger.close_loups],
            [call.args[0] for call in ca_patch.call_args_list])
        ca_patch.reset_mock()
        # heure / heure_chain
        self.assertFalse(bdd.Tache.query.all())

        # qui = "loups", heure = 15h12, heure_chain = None
        ctx = mock_discord.get_ctx(close, "loups", heure="15h12")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        ca_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open loups")
        bdd.Tache.delete(*taches)

        # qui = "loups", heure = 15h12, heure_chain = 7h
        ctx = mock_discord.get_ctx(close, "loups", heure="15h12",
                                   heure_chain="7h")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.reset_mock(return_value=True)
        ca_patch.reset_mock()
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 1)
        self.assertEqual(taches[0].commande, "!open loups 7h 15h12")
        bdd.Tache.delete(*taches)

        # ---- action ----
        bdd.BaseAction(slug="ouiz", trigger_debut=bdd.ActionTrigger.temporel,
                       heure_debut=datetime.time(15, 12)).add()
        action1 = bdd.Action(_base_slug="ouiz", joueur=joueur1)
        action2 = bdd.Action(_base_slug="ouiz", joueur=joueur1)
        action3 = bdd.Action(_base_slug="ouiz", joueur=joueur2)
        bdd.Action.add(action1, action2, action3)

        # qui = "action", heure = "15h12"
        ctx = mock_discord.get_ctx(close, "action", "15h12")
        rj_patch.return_value = {joueur1: [action1, action2],
                                 joueur2: [action3]}
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("close", "action", "15h12")
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Fermeture actions
        self.assertEqual(ca_patch.call_count, 3)
        self.assertEqual(
            [action1, action2, action3],
            [call.args[0] for call in ca_patch.call_args_list])
        ca_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock_discord.interact()
    @mock.patch("lgrez.features.open_close.recup_joueurs")  # tested before
    async def test_remind(self, rj_patch):
        """Unit tests for !remind command."""
        # async def remind(self, ctx, qui, heure=None)
        remind = self.cog.remind
        mock_bdd.add_campsroles()
        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                             votant_village=True, votant_loups=True,
                             vote_condamne_="non défini", vote_maire_="non défini",
                             vote_loups_="non défini")
        joueur2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                             votant_village=True, votant_loups=True,
                             vote_condamne_="non défini", vote_maire_="non défini",
                             vote_loups_="non défini")
        joueur3 = bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3")
        joueurs = [joueur1, joueur2, joueur3]
        bdd.Joueur.add(*joueurs)
        bases = [bdd.BaseAction(slug=trigger.name, trigger_debut=trigger)
                 for trigger in bdd.ActionTrigger]
        ac_oc = [bdd.Action(joueur=joueur1, base=base)
                 for joueur in joueurs for base in bases]

        # bad qui
        ctx = mock_discord.get_ctx(remind, "bzeepzopsl")
        rj_patch.side_effect = ValueError
        with self.assertRaises(ValueError):
            await ctx.invoke()
        rj_patch.assert_called_once_with("remind", "bzeepzopsl", None)
        rj_patch.reset_mock(side_effect=True)

        # qui = "cond", heure = None
        ctx = mock_discord.get_ctx(remind, "cond")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("remind", "cond", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("pour voter pour le condamné",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("pour voter pour le condamné",
                      chan2.send.call_args.args[0])

        # ---- maire ----
        # qui = "maire", heure = None
        ctx = mock_discord.get_ctx(remind, "maire")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("remind", "maire", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("pour élire le nouveau maire",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("pour élire le nouveau maire",
                      chan2.send.call_args.args[0])

        # ---- loups ----
        # qui = "loups", heure = None
        ctx = mock_discord.get_ctx(remind, "loups")
        rj_patch.return_value = [joueur1, joueur2]
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("remind", "loups", None)
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Information joueurs
        chan1.send.assert_called_once()
        self.assertIn("pour voter pour la victime",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("pour voter pour la victime",
                      chan2.send.call_args.args[0])

        # ---- action ----
        bdd.BaseAction(slug="ouiz", trigger_debut=bdd.ActionTrigger.temporel,
                       heure_debut=datetime.time(15, 12)).add()
        action1 = bdd.Action(_base_slug="ouiz", joueur=joueur1)
        action2 = bdd.Action(_base_slug="ouiz", joueur=joueur1)
        action3 = bdd.Action(_base_slug="ouiz", joueur=joueur2)
        bdd.Action.add(action1, action2, action3)

        # qui = "action", heure = "15h12"
        ctx = mock_discord.get_ctx(remind, "action", "15h12")
        rj_patch.return_value = {joueur1: [action1, action2],
                                 joueur2: [action3]}
        with mock_discord.mock_members_and_chans(joueur1, joueur2):
            chan1, chan2 = joueur1.private_chan, joueur2.private_chan
            await ctx.invoke()
        rj_patch.assert_called_once_with("remind", "action", "15h12")
        rj_patch.reset_mock(return_value=True)
        # Liste joueurs concernés
        ctx.send.assert_called_once()
        self.assertIn("Joueur1", ctx.send.call_args.args[0])
        self.assertIn("Joueur2", ctx.send.call_args.args[0])
        self.assertNotIn("Joueur3", ctx.send.call_args.args[0])
        # Information joueurs
        self.assertEqual(chan1.send.call_count, 2)
        self.assertIn("pour utiliser ton action",
                      chan1.send.call_args.args[0])
        chan2.send.assert_called_once()
        self.assertIn("pour utiliser ton action",
                      chan2.send.call_args.args[0])


    @mock_bdd.patch_db      # Empty database for this method
    @mock_discord.interact()
    @mock.patch("lgrez.features.open_close._do_refill")  # tested before
    async def test_refill(self, dr_patch):
        """Unit tests for !refill command."""
        # async def refill(self, ctx, motif, *, cible=None)
        refill = self.cog.refill

        config.refills_full = ["uikz"]
        config.refills_one = ["forgebonk", "rebootax", "divvvvvvin"]
        config.refills_divins = ["divvvvvvin"]

        mock_bdd.add_campsroles()
        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        joueur2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2")
        joueurs = [joueur1, joueur2]
        bdd.Joueur.add(*joueurs)
        nondivin = set(config.refills_full
                       + config.refills_one) ^ set(config.refills_divins)
        bdd.BaseAction(slug="baz_all", refill=", ".join(nondivin)).add()
        act_all_refs = bdd.Action(joueur=joueur1, _base_slug="baz_all",
                                  charges=3)
        actions = {}
        for kw in [*config.refills_full, *config.refills_one]:
            base = f"baz_{kw}"
            bdd.BaseAction(slug=base, refill=kw).add()
            actions[kw] = [
                bdd.Action(joueur=joueur1, _base_slug=base, charges=3),
                bdd.Action(joueur=joueur2, _base_slug=base, charges=3),
                act_all_refs,
            ]
        action_none = bdd.Action(joueur=joueur1, _base_slug=base, charges=None)
        all_actions = {ac for acs in actions.values() for ac in acs}
        bdd.Action.add(*all_actions, action_none)

        # bad motif
        ctx = mock_discord.get_ctx(refill, "bzeepzopsl")
        await ctx.invoke()
        ctx.send.assert_called_once()
        self.assertIn("pas un motif valide", ctx.send.call_args.args[0])
        dr_patch.assert_not_called()

        # motif = divin, cible = "all", abort
        ctx = mock_discord.get_ctx(refill, "divvvvvvin", cible="all")
        with mock_discord.interact(("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent(
            "es-tu sûr ?",
            "Mission aborted",
        )
        dr_patch.assert_not_called()

        # motif = divin, cible = "all", proceed
        ctx = mock_discord.get_ctx(refill, "divvvvvvin", cible="all")
        with mock_discord.interact(("yes_no", True)):
            await ctx.invoke()
        ctx.assert_sent(
            "es-tu sûr ?",
            "répondant aux critères",
        )
        dr_patch.assert_called_once()
        motif, refilled = dr_patch.call_args.args
        self.assertEqual("divvvvvvin", motif)
        self.assertEqual(set(all_actions), set(refilled))
        dr_patch.reset_mock()

        # motif = divin, cible = Joueur1
        ctx = mock_discord.get_ctx(refill, "divvvvvvin", cible="Joueur1")
        await ctx.invoke()
        ctx.assert_sent("répondant aux critères")
        dr_patch.assert_called_once()
        motif, refilled = dr_patch.call_args.args
        self.assertEqual("divvvvvvin", motif)
        self.assertEqual(set(ac for ac in all_actions if ac.joueur == joueur1),
                         set(refilled))
        dr_patch.reset_mock()

        # motif = autre, cible = "all"
        for motif in ["uikz", "forgebonk", "rebootax"]:
            ctx = mock_discord.get_ctx(refill, motif, cible="all")
            await ctx.invoke()
            ctx.assert_sent("répondant aux critères")
            dr_patch.assert_called_once()
            motifc, refilled = dr_patch.call_args.args
            self.assertEqual(motif, motifc)
            self.assertEqual(set(actions[motif]), set(refilled))
            dr_patch.reset_mock()

        # motif = autre, cible = Joueur1
        for motif in ["uikz", "forgebonk", "rebootax"]:
            ctx = mock_discord.get_ctx(refill, motif, cible="Joueur1")
            await ctx.invoke()
            ctx.assert_sent("répondant aux critères")
            dr_patch.assert_called_once()
            motifc, refilled = dr_patch.call_args.args
            self.assertEqual(motif, motifc)
            self.assertEqual(set(ac for ac in actions[motif]
                                 if ac.joueur == joueur1),
                             set(refilled))
            dr_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock_discord.interact()
    async def test_cparti(self):
        """Unit tests for !cparti command."""
        # async def cparti(self, ctx)
        cparti = self.cog.cparti

        mock_bdd.add_campsroles()
        joueur1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        joueur2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2")
        joueurs = [joueur1, joueur2]
        bdd.Joueur.add(*joueurs)
        bas = [
            bdd.BaseAction(slug="baz_strt", trigger_debut="start"),
            bdd.BaseAction(slug="baz_prm", trigger_debut="perma"),
            bdd.BaseAction(slug="baz_aut", trigger_debut="temporel")
        ]
        bdd.BaseAction.add(*bas)
        actions = [*[bdd.Action(base=base, joueur=joueur1) for base in bas],
                   *[bdd.Action(base=base, joueur=joueur2) for base in bas]]
        bdd.Action.add(*actions)

        # abort 1
        ctx = mock_discord.get_ctx(cparti)
        with mock_discord.interact(("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent(
            "",
            "Mission aborted",
        )

        # abort 2
        ctx = mock_discord.get_ctx(cparti)
        with mock_discord.interact(("yes_no", True), ("yes_no", False)):
            await ctx.invoke()
        ctx.assert_sent(
            "",
            "",
            "Mission aborted",
        )

        # proceed
        ctx = mock_discord.get_ctx(cparti)
        with mock_discord.interact(("yes_no", True), ("yes_no", True)):
            await ctx.invoke()
        ctx.assert_sent(
            "",
            "",
            "tout bon",
        )
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 8)
        commands_debuts = {
            " ".join(tache.commande.split(maxsplit=2)[:2]) for tache in taches
        }
        expected = {
            "!open cond",
            "!open maire",
            "!open loups",
            "!send all",
            *{f"!open {action.id}" for action in actions
              if action.base.slug in ("baz_strt", "baz_prm")}
        }
        self.assertEqual(commands_debuts, expected)



"""lg-rez / features / Commandes informatives

Commandes donnant aux joueurs des informations sur le jeu, leurs
actions, les joueurs en vie et morts...

"""

from discord.ext import commands

from lgrez import config
from lgrez.blocs import tools
from lgrez.bdd import (Joueur, Role, Camp, Action, BaseAction,
                       ActionTrigger, Vote)
from lgrez.features import gestion_actions


def _roles_list(roles):
    return "\n".join(
        str(role.camp.discord_emoji_or_none or "")
        + tools.code(f"{role.nom.ljust(25)} {role.description_courte}")
        for role in roles if not role.nom.startswith("(")
    )


class Informations(commands.Cog):
    """Commandes pour en savoir plus sur soi et les autres"""

    @commands.command(aliases=["role", "rôles", "rôle"])
    async def roles(self, ctx, *, role=None):
        """Affiche la liste des rôles / des informations sur un rôle

        Args:
            role: le nom d'un rôle, pour les informations sur ce rôle.

        Sans argument, liste tous les rôles existants.
        Voir aussi la commande `!camps`.
        """
        if role:
            role = tools.remove_accents(role.lower())
            role = role.strip("<>[](){}")

        if not role:
            roles = Role.query.filter_by(actif=True).order_by(Role.nom).all()
        else:
            roles = Role.find_nearest(role, col=Role.nom)
            if not roles:
                await ctx.send(f"Rôle \"{role}\" non trouvé.")
                return

            await ctx.send(embed=roles[0][0].embed)
            return

        await tools.send_blocs(
            ctx,
            f"Rôles trouvés :\n{_roles_list(roles)}"
            + "\n" + tools.ital(f"({tools.code('!role <role>')} "
                                "pour plus d'informations sur un rôle.)")
        )


    @commands.command(aliases=["camp"])
    async def camps(self, ctx, *, camp=None):
        """Affiche la liste des camps / les rôles d'un camp

        Args:
            camp: le nom d'un camp, pour les informations sur ce camp.

        Sans argument, liste tous les camps existants.
        Voir aussi la commande `!role`.
        """
        if camp:
            camp = tools.remove_accents(camp.lower())
            camp = camp.strip("<>[](){}")

        if not camp:
            camps = Camp.query.filter_by(public=True).order_by(Camp.nom).all()
        else:
            camps = Camp.find_nearest(camp, col=Camp.nom)
            if not camps:
                await ctx.send(f"Camp \"{camp}\" non trouvé.")
                return

            await ctx.send(embed=camps[0][0].embed)
            await tools.send_blocs(
                ctx,
                f"Rôles dans ce camp :\n{_roles_list(camps[0][0].roles)}"
                + "\n" + tools.ital(f"({tools.code('!role <role>')} "
                                    "pour plus d'informations sur un rôle.)")
            )
            return

        await tools.send_blocs(
            ctx,
            "Camps trouvés :\n"
            + "\n".join(str(camp.discord_emoji_or_none or "") + " " + camp.nom
                        for camp in camps if not camp.nom.startswith("("))
            + "\n" + tools.ital(f"({tools.code('!camp <camp>')} "
                                "pour plus d'informations sur un camp.)")
        )


    @commands.command()
    @tools.mjs_only
    async def rolede(self, ctx, *, cible=None):
        """Donne le rôle d'un joueur (COMMANDE MJ)

        Args:
            cible: le joueur dont on veut connaître le rôle
        """
        joueur = await tools.boucle_query_joueur(ctx, cible, "Qui ?")
        if joueur:
            await ctx.send(joueur.role.nom_complet)


    @commands.command()
    @tools.mjs_only
    async def quiest(self, ctx, *, nomrole):
        """Liste les joueurs ayant un rôle donné (COMMANDE MJ)

        Args:
            nomrole: le rôle qu'on cherche (doit être un slug ou un nom
                de rôle valide)
        """
        roles = Role.find_nearest(nomrole)
        if roles:
            role = roles[0][0]
        else:
            roles = Role.find_nearest(nomrole, col=Role.nom)
            if roles:
                role = roles[0][0]
            else:
                await ctx.send("Connais pas")
                return

        joueurs = Joueur.query.filter_by(role=role).\
                               filter(Joueur.est_vivant).all()
        await ctx.send(
            f"{role.nom_complet} : "
            + (", ".join(joueur.nom for joueur in joueurs)
               if joueurs else "Personne.")
        )


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def menu(self, ctx):
        """Affiche des informations et boutons sur les votes / actions en cours

        Le menu a une place beaucoup moins importante ici que sur
        Messenger, vu que tout est accessible par commandes.
        """
        member = ctx.author
        joueur = Joueur.from_member(member)

        reacts = []
        r = "––– MENU –––\n\n"

        try:
            vaction = joueur.action_vote(Vote.cond)
        except RuntimeError:
            await ctx.send("Minute papillon, le jeu n'est pas encore lancé !")
            return

        if vaction.is_open:
            r += (f" - {config.Emoji.bucher}  Vote pour le bûcher en cours – "
                  f"vote actuel : {tools.code(vaction.decision)}\n")
            reacts.append(config.Emoji.bucher)

        vaction = joueur.action_vote(Vote.maire)
        if vaction.is_open:
            r += (f" - {config.Emoji.maire}  Vote pour le maire en cours – "
                  f"vote actuel : {tools.code(vaction.decision)}\n")
            reacts.append(config.Emoji.maire)

        vaction = joueur.action_vote(Vote.loups)
        if vaction.is_open:
            r += (f" - {config.Emoji.lune}  Vote des loups en cours – "
                  f"vote actuel : {tools.code(vaction.decision)}\n")
            reacts.append(config.Emoji.lune)

        if not reacts:
            r += "Aucun vote en cours.\n"

        actions = [ac for ac in joueur.actions_actives if ac.is_open]
        if actions:
            for action in actions:
                r += (f" - {config.Emoji.action}  Action en cours : "
                      f"{tools.code(action.base.slug)} (id {action.id}) – "
                      f"décision : {tools.code(action.decision)}\n")
            reacts.append(config.Emoji.action)
        else:
            r += "Aucune action en cours.\n"

        message = await ctx.send(
            r + f"\n{tools.code('!infos')} pour voir ton rôle et tes "
            f"actions, {tools.code('@MJ')} en cas de problème"
        )
        for react in reacts:
            await message.add_reaction(react)


    @commands.command()
    @tools.vivants_only
    @tools.private
    async def infos(self, ctx):
        """Affiche tes informations de rôle / actions

        Toutes les actions liées à ton rôle (et parfois d'autres) sont
        indiquées, même celles que tu ne peux pas utiliser pour
        l'instant (plus de charges, déclenchées automatiquement...)
        """
        member = ctx.author
        joueur = Joueur.from_member(member)
        r = ""

        r += f"Ton rôle actuel : {tools.bold(joueur.role.nom_complet)}\n"
        r += tools.ital(f"({tools.code(f'!roles {joueur.role.slug}')} "
                        "pour tout savoir sur ce rôle)")

        if joueur.actions_actives:
            r += "\n\nActions :"
            r += tools.code_bloc("\n".join((
                f" - {action.base.slug.ljust(20)} "
                + (f"Cooldown : {action.cooldown}" if action.cooldown
                   else action.base.temporalite).ljust(22)
                + (f"   {action.charges} charge(s)"
                    + (" pour cette semaine"
                        if "weekends" in action.base.refill else "")
                    if isinstance(action.charges, int) else "Illimitée")
            ) for action in joueur.actions_actives))
            # Vraiment désolé pour cette immondice j'ai la flemme
        else:
            r += "\n\nAucune action disponible."

        await ctx.send(
            f"{r}\n{tools.code('!menu')} pour voir les votes et "
            f"actions en cours, {tools.code('@MJ')} en cas de problème"
        )


    @commands.command()
    @tools.mjs_only
    async def actions(self, ctx, *, cible=None):
        """Affiche et modifie les actions d'un joueur (COMMANDE MJ)

        Args:
            cible: le joueur dont on veut voir ou modifier les actions

        Warning:
            Commande expérimentale, non testée.
        """
        joueur = await tools.boucle_query_joueur(ctx, cible, "Qui ?")
        actions = [ac for ac in joueur.actions if not ac.vote]

        r = f"Rôle : {joueur.role.nom_complet or joueur.role}\n"

        # if not actions:
        #     r += "Aucune action pour ce joueur."
        #     await ctx.send(r)
        #     return

        r += "Actions :"
        r += tools.code_bloc(
            "#️⃣  id   active  baseaction               début"
            "     fin       cd   charges   refill     \n"
            "---------------------------------------------"
            "---------------------------------------------\n"
            + "\n".join([(
                tools.emoji_chiffre(i + 1) + "  "
                + str(action.id).ljust(5)
                + str(action.active).ljust(8)
                + action.base.slug.ljust(25)
                + str(action.base.heure_debut
                      if action.base.trigger_debut == ActionTrigger.temporel
                      else action.base.trigger_debut).ljust(10)
                + str(action.base.heure_fin
                      if action.base.trigger_fin == ActionTrigger.temporel
                      else action.base.trigger_fin).ljust(10)
                + str(action.cooldown).ljust(5)
                + str(action.charges).ljust(10)
                + str(action.base.refill)
            ) for i, action in enumerate(actions)])
        )
        r += "Modifier/ajouter/stop :"
        message = await ctx.send(r)
        i = await tools.choice(message, len(actions),
                               additionnal={"🆕": -1, "⏹": 0})

        if i < 0:     # ajouter
            await ctx.send("Slug de la baseaction à ajouter ? "
                           "(voir Gsheet Rôles et actions)")
            base = await tools.boucle_query(ctx, BaseAction)

            await ctx.send("Cooldown ? (nombre entier)")
            mess = await tools.wait_for_message_here(ctx)
            cooldown = int(mess.content)

            await ctx.send(f"Charges ? ({tools.code('None')} pour illimité)")
            mess = await tools.wait_for_message_here(ctx)
            charges = int(mess.content) if mess.content.isdigit() else None

            action = Action(joueur=joueur, base=base, cooldown=cooldown,
                            charges=charges)
            gestion_actions.add_action(action)
            await ctx.send(f"Action ajoutée (id {action.id}).")
            return

        elif i == 0:
            await ctx.send("Au revoir.")
            return

        # Modifier
        action = actions[i - 1]
        stop = False
        while not stop:
            await ctx.send(
                "Modifier : (parmi `active, cd, charges`}) ; `valider` pour "
                "finir ; `supprimer` pour supprimer l'action.\n(Pour modifier "
                "les attributs de la baseaction, modifier le Gsheet et "
                "utiliser `!fillroles` ; pour ouvrir/fermer l'action, "
                f"utiliser `!open {action.id}` / `!close {action.id}`.)"
            )
            mess = await tools.wait_for_message_here(ctx)
            modif = mess.content.lower()

            if modif == "active":
                mess = await ctx.send("Action active ?")
                action.active = await tools.yes_no(mess)

            elif modif in ["cd", "cooldown"]:
                await ctx.send("Combien ?")
                mess = await tools.wait_for_message_here(ctx)
                action.cooldown = int(mess.content)

            elif modif == "charges":
                await ctx.send("Combien ? (`None` pour illimité)")
                mess = await tools.wait_for_message_here(ctx)
                action.charges = (int(mess.content)
                                  if mess.content.isdigit() else None)

            elif modif == "valider":
                action.update()
                await ctx.send("Fait.")
                stop = True

            elif modif == "supprimer":
                mess = await ctx.send("Supprimer l'action ? (privilégier "
                                      "l'archivage  `active = False`)")
                if await tools.yes_no(mess):
                    action.delete()
                    await ctx.send("Fait.")
                    stop = True

            else:
                await ctx.send("Valeur incorrecte")


    @commands.command(aliases=["joueurs", "vivant"])
    async def vivants(self, ctx):
        """Affiche la liste des joueurs vivants

        Aussi dite : « liste des joueurs qui seront bientôt morts »
        """
        joueurs = Joueur.query.filter(Joueur.est_vivant).\
                               order_by(Joueur.nom).all()

        mess = " Joueur                     en chambre\n"
        mess += "––––––––––––––––––––––––––––––––––––––––––––––\n"
        if config.demande_chambre:
            for joueur in joueurs:
                mess += f" {joueur.nom.ljust(25)}  {joueur.chambre}\n"
        else:
            for joueur in joueurs:
                mess += f" {joueur.nom}\n"

        await tools.send_code_blocs(
            ctx, mess, prefixe=f"Les {len(joueurs)} joueurs vivants sont :"
        )


    @commands.command(aliases=["mort"])
    async def morts(self, ctx):
        """Affiche la liste des joueurs morts

        Aussi dite : « liste des joueurs qui mangent leurs morts »
        """
        joueurs = Joueur.query.filter(Joueur.est_mort).\
                               order_by(Joueur.nom).all()

        if joueurs:
            mess = ""
            for joueur in joueurs:
                mess += f" {joueur.nom}\n"
        else:
            mess = "Toi (mais tu ne le sais pas encore)"

        await tools.send_code_blocs(
            ctx, mess, prefixe=f"Les {len(joueurs) or ''} morts sont :"
        )



import datetime
import unittest
from unittest import mock

import discord
import freezegun

from lgrez import config, bdd
from lgrez.features import gestion_actions
from lgrez.blocs import gsheets
from test import mock_discord, mock_bdd, mock_env


def base_joueurs():
    return [
        bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1", chambre="Ch1",
                   statut=bdd.Statut.vivant, role=bdd.Role.default(),
                   camp=bdd.Camp.default(), votant_village=False,
                   votant_loups=False, role_actif=False),
        bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2", chambre="Ch2",
                   statut=bdd.Statut.mort, _role_slug="role2",
                   _camp_slug="camp2", votant_village=True,
                   votant_loups=False, role_actif=False),
        bdd.Joueur(discord_id=3, chan_id_=31, nom="Joueur3", chambre="Ch3",
                   statut=bdd.Statut.MV, _role_slug="role3",
                   _camp_slug="camp3", votant_village=True,
                   votant_loups=True, role_actif=False),
        bdd.Joueur(discord_id=4, chan_id_=41, nom="Joueur4", chambre="Ch4",
                   statut=bdd.Statut.immortel, _role_slug="role4",
                   _camp_slug="camp4", votant_village=True,
                   votant_loups=True, role_actif=True),
    ]


class TestGestionActionsFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.features.gestion_actions functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()


    @mock_bdd.patch_db      # Empty database for this method
    def test_add_action(self):
        """Unit tests for gestion_actions.add_action function."""
        # def add_action(action)
        add_action = gestion_actions.add_action
        mock_bdd.add_campsroles()
        j1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        j1.add()

        # basic
        ba1 = bdd.BaseAction(slug="ba1", trigger_debut=bdd.ActionTrigger.mort)
        ba1.add()
        act1 = bdd.Action(base=ba1, joueur=j1)
        add_action(act1)
        self.assertEqual(bdd.Action.query.all(), [act1])
        self.assertEqual(bdd.Tache.query.all(), [])
        act1.delete()

        # temporel
        ba2 = bdd.BaseAction(slug="ba2",
                             trigger_debut=bdd.ActionTrigger.temporel,
                             heure_debut=datetime.time(15, 2))
        ba2.add()
        act2 = bdd.Action(id=25, base=ba2, joueur=j1)
        add_action(act2)
        self.assertEqual(bdd.Action.query.all(), [act2])
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp.time(), datetime.time(15, 2))
        self.assertEqual(tache.commande, "!open 25")
        self.assertEqual(tache.action, act2)
        act2.delete()
        tache.delete()

        # perma
        ba3 = bdd.BaseAction(slug="ba3", trigger_debut=bdd.ActionTrigger.perma)
        ba3.add()
        act3 = bdd.Action(id=23, base=ba3, joueur=j1)
        add_action(act3)
        self.assertEqual(bdd.Action.query.all(), [act3])
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertAlmostEqual(tache.timestamp, datetime.datetime.now(),
                               delta=datetime.timedelta(seconds=1))
        self.assertEqual(tache.commande, "!open 23")
        self.assertEqual(tache.action, act3)


    @mock_bdd.patch_db      # Empty database for this method
    def test_delete_action(self):
        """Unit tests for gestion_actions.delete_action function."""
        # def delete_action(action)
        delete_action = gestion_actions.delete_action
        mock_bdd.add_campsroles()
        j1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        j1.add()

        # no taches
        ba1 = bdd.BaseAction(slug="ba1")
        ba1.add()
        act1 = bdd.Action(base=ba1, joueur=j1)
        act1.add()
        delete_action(act1)
        self.assertEqual(bdd.Action.query.all(), [])
        self.assertEqual(bdd.Tache.query.all(), [])

        # some taches
        act1 = bdd.Action(base=ba1, joueur=j1)
        act1.add()
        bdd.Tache(timestamp=datetime.datetime.now(), commande="t1",
                  action=act1).add()
        bdd.Tache(timestamp=datetime.datetime(2021, 3, 15, 15, 2, 0),
                  commande="t2", action=act1).add()
        delete_action(act1)
        self.assertEqual(bdd.Action.query.all(), [])
        self.assertEqual(bdd.Tache.query.all(), [])


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.gestion_actions.close_action")
    @mock.patch("lgrez.blocs.tools.log")
    async def test_open_action(self, log_patch, ca_patch):
        """Unit tests for gestion_actions.open_action function."""
        # async def open_action(action)
        open_action = gestion_actions.open_action
        mock_bdd.add_campsroles()
        j1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1",
                        role_actif=True)
        j1.add()

        # test préliminaire 1 - en cooldown, non temporel
        ba1 = bdd.BaseAction(slug="ba1")
        ba1.add()
        act1 = bdd.Action(base=ba1, joueur=j1, cooldown=3)
        act1.add()
        with mock_discord.mock_members_and_chans(j1):
            await open_action(act1)
        self.assertEqual(act1.cooldown, 2)
        log_patch.assert_called_once()
        self.assertIn(repr(act1), log_patch.call_args.args[0])
        self.assertIn("en cooldown", log_patch.call_args.args[0])
        self.assertEqual(bdd.Tache.query.all(), [])
        ca_patch.assert_not_called()
        log_patch.reset_mock()

        # test préliminaire 2 - en cooldown, temporel
        ba2 = bdd.BaseAction(slug="ba2",
                             trigger_debut=bdd.ActionTrigger.temporel,
                             heure_debut=datetime.time(15, 2))
        ba2.add()
        act2 = bdd.Action(id=23, base=ba2, joueur=j1, cooldown=3)
        act2.add()
        with mock_discord.mock_members_and_chans(j1):
            await open_action(act2)
        self.assertEqual(act2.cooldown, 2)
        log_patch.assert_called_once()
        self.assertIn(repr(act2), log_patch.call_args.args[0])
        self.assertIn("en cooldown", log_patch.call_args.args[0])
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp.time(), datetime.time(15, 2))
        self.assertEqual(tache.commande, "!open 23")
        self.assertEqual(tache.action, act2)
        log_patch.reset_mock()
        ca_patch.assert_not_called()
        tache.delete()

        # test préliminaire 3 - role non actif, non temporel
        j2 = bdd.Joueur(discord_id=2, chan_id_=21, nom="Joueur2",
                        role_actif=False)
        j2.add()
        act3 = bdd.Action(id=24, base=ba1, joueur=j2, cooldown=0)
        act3.add()
        with mock_discord.mock_members_and_chans(j2):
            await open_action(act3)
        self.assertEqual(act3.cooldown, 0)
        log_patch.assert_called_once()
        self.assertIn(repr(act3), log_patch.call_args.args[0])
        self.assertIn("role_actif == False", log_patch.call_args.args[0])
        self.assertEqual(bdd.Tache.query.all(), [])
        ca_patch.assert_not_called()
        log_patch.reset_mock()

        # test préliminaire 4 - role non actif, temporel
        act4 = bdd.Action(id=25, base=ba2, joueur=j2, cooldown=0)
        act4.add()
        with mock_discord.mock_members_and_chans(j2):
            await open_action(act4)
        self.assertEqual(act4.cooldown, 0)
        log_patch.assert_called_once()
        self.assertIn(repr(act4), log_patch.call_args.args[0])
        self.assertIn("role_actif == False", log_patch.call_args.args[0])
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp.time(), datetime.time(15, 2))
        self.assertEqual(tache.commande, "!open 25")
        self.assertEqual(tache.action, act4)
        log_patch.reset_mock()
        ca_patch.assert_not_called()
        tache.delete()

        # test préliminaire 5 - action automatique, non temporelle
        ba3 = bdd.BaseAction(slug="ba3",
                             trigger_debut=bdd.ActionTrigger.perma,
                             trigger_fin=bdd.ActionTrigger.auto)
        ba3.add()
        act5 = bdd.Action(id=27, base=ba3, joueur=j1, cooldown=0)
        act5.add()
        with mock_discord.mock_members_and_chans(j1):
            await open_action(act5)
        self.assertEqual(act5.cooldown, 0)
        log_patch.assert_called_once()
        self.assertIn(repr(act5), log_patch.call_args.args[0])
        self.assertIn("automatique", log_patch.call_args.args[0])
        self.assertEqual(bdd.Tache.query.all(), [])
        ca_patch.assert_called_once_with(act5)
        log_patch.reset_mock()
        ca_patch.reset_mock()

        # test préliminaire 6 - action automatique, temporelle
        ba4 = bdd.BaseAction(slug="ba4",
                             trigger_debut=bdd.ActionTrigger.temporel,
                             heure_debut=datetime.time(15, 2),
                             trigger_fin=bdd.ActionTrigger.auto)
        ba4.add()
        act6 = bdd.Action(id=28, base=ba4, joueur=j1, cooldown=0)
        act6.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act6)
        self.assertEqual(act6.cooldown, 0)
        log_patch.assert_called_once()
        self.assertIn("ba4", log_patch.call_args.args[0])
        self.assertIn("Joueur1", log_patch.call_args.args[0])
        self.assertIn("pas vraiment automatique", log_patch.call_args.args[0])
        self.assertIn(str(config.Role.mj.mention), log_patch.call_args.args[0])
        self.assertIn(str(chan.mention), log_patch.call_args.args[0])
        self.assertEqual(bdd.Tache.query.all(), [])
        ca_patch.assert_called_once_with(act6)
        log_patch.reset_mock()
        ca_patch.reset_mock()


        # --- fin des tests préliminaires
        # fin hors temp/delta/perma, déjà ouverte
        ba5 = bdd.BaseAction(slug="ba5",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.close_cond)
        ba5.add()
        act7 = bdd.Action(base=ba5, joueur=j1, decision_="rien")
        act7.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act7)
        mock_discord.assert_sent(chan, [
            "tu peux utiliser quand tu le souhaites", "ba5", "!action",
        ])
        mock_discord.assert_not_sent(chan, "tu as jusqu'à")
        self.assertEqual(act7.decision_, "rien")
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        self.assertEqual(bdd.Tache.query.all(), [])
        log_patch.assert_not_called()
        ca_patch.assert_not_called()

        # fin hors temp/delta/perma, fermée
        ba6 = bdd.BaseAction(slug="ba6",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.close_cond)
        ba6.add()
        act8 = bdd.Action(id=30, base=ba6, joueur=j1, decision_=None)
        act8.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act8)
        mock_discord.assert_sent(chan, [
            "Tu peux maintenant utiliser", "ba6", "!action",
        ])
        mock_discord.assert_not_sent(chan, "tu as jusqu'à")
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        self.assertEqual(act8.decision_, "rien")
        self.assertEqual(bdd.Tache.query.all(), [])
        log_patch.assert_not_called()
        ca_patch.assert_not_called()


        # fin temp, déjà ouverte
        ba7 = bdd.BaseAction(slug="ba7",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.temporel,
                             heure_fin=datetime.time(15, 2))
        ba7.add()
        act9 = bdd.Action(id=31, base=ba7, joueur=j1, decision_="rien")
        act9.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act9)
        mock_discord.assert_sent(chan, [
            "tu peux utiliser quand tu le souhaites", "ba7", "!action",
            f"Tu as jusqu'à {datetime.time(15, 2)}"
        ])
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        self.assertEqual(act9.decision_, "rien")
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual([tache.action for tache in taches], [act9]*2)
        self.assertEqual(
            {(tache.commande, tache.timestamp.time()) for tache in taches},
            {("!close 31", datetime.time(15, 2)),
             ("!remind 31", datetime.time(14, 32))}
        )
        bdd.Tache.delete(*taches)
        log_patch.assert_not_called()
        ca_patch.assert_not_called()

        # fin temp, fermée
        ba8 = bdd.BaseAction(slug="ba8",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.temporel,
                             heure_fin=datetime.time(15, 2))
        ba8.add()
        act10 = bdd.Action(id=32, base=ba8, joueur=j1, decision_=None)
        act10.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act10)
        mock_discord.assert_sent(chan, [
            "Tu peux maintenant utiliser", "ba8", "!action",
            f"Tu as jusqu'à {datetime.time(15, 2)}"
        ])
        self.assertEqual(act10.decision_, "rien")
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual([tache.action for tache in taches], [act10]*2)
        self.assertEqual(
            {(tache.commande, tache.timestamp.time()) for tache in taches},
            {("!close 32", datetime.time(15, 2)),
             ("!remind 32", datetime.time(14, 32))}
        )
        bdd.Tache.delete(*taches)
        log_patch.assert_not_called()
        ca_patch.assert_not_called()


        # fin delta (fermée)
        ba9 = bdd.BaseAction(slug="ba9",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.delta,
                             heure_fin=datetime.time(1, 10, 7))
        ba9.add()
        act11 = bdd.Action(id=33, base=ba9, joueur=j1, decision_=None)
        act11.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            with freezegun.freeze_time(datetime.datetime(1, 1, 1, 16, 28, 4)):
                await open_action(act11)
        mock_discord.assert_sent(chan, [
            "Tu peux maintenant utiliser", "ba9", "!action",
            f"Tu as jusqu'à {datetime.time(17, 38, 11)}"
        ])
        self.assertEqual(act11.decision_, "rien")
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        taches = bdd.Tache.query.all()
        self.assertEqual(len(taches), 2)
        self.assertEqual([tache.action for tache in taches], [act11]*2)
        self.assertEqual(
            {(tache.commande, tache.timestamp.time()) for tache in taches},
            {("!close 33", datetime.time(17, 38, 11)),
             ("!remind 33", datetime.time(17, 8, 11))}
        )
        bdd.Tache.delete(*taches)
        log_patch.assert_not_called()
        ca_patch.assert_not_called()


        # fin perma (fermée) - WE d'abord
        ba10 = bdd.BaseAction(slug="ba10",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.perma)
        ba10.add()
        act12 = bdd.Action(id=34, base=ba10, joueur=j1, decision_=None)
        act12.add()
        with mock.patch("lgrez.blocs.tools.next_occurence",
                        return_value=datetime.datetime(1, 1, 1, 16, 28, 4)), \
             mock.patch("lgrez.blocs.tools.debut_pause",
                        return_value=datetime.datetime(1, 1, 1, 15, 28, 4)), \
             mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act12)
        mock_discord.assert_sent(chan, [
            "Tu peux maintenant utiliser", "ba10", "!action",
        ])
        self.assertEqual(act12.decision_, "rien")
        mock_discord.assert_not_sent(chan, "Tu as jusqu'à",)
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.action, act12)
        self.assertEqual(tache.commande, "!close 34")
        self.assertEqual(tache.timestamp.time(), datetime.time(15, 28, 4))
        tache.delete()
        log_patch.assert_not_called()
        ca_patch.assert_not_called()

        # fin perma (fermée) - prochaine ouverture d'abord
        ba11 = bdd.BaseAction(slug="ba11",
                             trigger_debut=bdd.ActionTrigger.open_cond,
                             trigger_fin=bdd.ActionTrigger.perma)
        ba11.add()
        act13 = bdd.Action(id=35, base=ba11, joueur=j1, decision_=None)
        act13.add()
        with mock.patch("lgrez.blocs.tools.next_occurence",
                        return_value=datetime.datetime(1, 1, 1, 16, 28, 4)), \
             mock.patch("lgrez.blocs.tools.debut_pause",
                        return_value=datetime.datetime(1, 1, 1, 17, 28, 4)), \
             mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await open_action(act13)
        mock_discord.assert_sent(chan, [
            "Tu peux maintenant utiliser", "ba11", "!action",
        ])
        self.assertEqual(act13.decision_, "rien")
        mock_discord.assert_not_sent(chan, "Tu as jusqu'à",)
        chan.send.return_value.add_reaction.assert_called_once_with(
            config.Emoji.action
        )
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.action, act13)
        self.assertEqual(tache.commande, "!open 35")
        self.assertEqual(tache.timestamp.time(), datetime.time(16, 28, 4))
        tache.delete()
        log_patch.assert_not_called()
        ca_patch.assert_not_called()


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.gestion_actions.delete_action")
    async def test_close_action(self, da_patch):
        """Unit tests for gestion_actions.close_action function."""
        # async def close_action(action)
        close_action = gestion_actions.close_action
        mock_bdd.add_campsroles()
        j1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        j1.add()

        # pas de décision, no base_cooldown, no temporel/perma
        ba1 = bdd.BaseAction(slug="ba1", instant=False, base_cooldown=0,
                             trigger_debut=bdd.ActionTrigger.open_cond)
        ba1.add()
        act1 = bdd.Action(base=ba1, joueur=j1, decision_="rien")
        act1.add()
        await close_action(act1)
        self.assertIsNone(act1.decision_)
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # pas de décision, base_cooldown = 4, no temporel/perma
        ba2 = bdd.BaseAction(slug="ba2", instant=False, base_cooldown=4,
                             trigger_debut=bdd.ActionTrigger.open_cond)
        ba2.add()
        act2 = bdd.Action(base=ba2, joueur=j1, cooldown=0, decision_="rien")
        act2.add()
        await close_action(act2)
        self.assertIsNone(act2.decision_)
        self.assertEqual(act2.cooldown, 4)
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # pas de décision, no base_cooldown, temporel
        ba3 = bdd.BaseAction(slug="ba3", instant=False, base_cooldown=0,
                             trigger_debut=bdd.ActionTrigger.temporel,
                             heure_debut=datetime.time(15, 2))
        ba3.add()
        act3 = bdd.Action(id=3, base=ba3, joueur=j1, decision_="rien")
        act3.add()
        with mock.patch("lgrez.blocs.tools.next_occurence") as no_patch:
            no_patch.return_value = datetime.datetime(1, 1, 1, 15, 2)
            await close_action(act3)
        no_patch.assert_called_once_with(datetime.time(15, 2))
        self.assertIsNone(act3.decision_)
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, no_patch.return_value)
        self.assertEqual(tache.commande, "!open 3")
        self.assertEqual(tache.action, act3)
        tache.delete()
        da_patch.assert_not_called()

        # pas de décision, no base_cooldown, perma
        ba4 = bdd.BaseAction(slug="ba4", instant=False, base_cooldown=0,
                             trigger_debut=bdd.ActionTrigger.perma)
        ba4.add()
        act4 = bdd.Action(id=4, base=ba4, joueur=j1, decision_="rien")
        act4.add()
        with mock.patch("lgrez.blocs.tools.fin_pause") as fp_patch:
            fp_patch.return_value = datetime.datetime(1, 1, 1, 10, 4)
            await close_action(act4)
        fp_patch.assert_called_once()
        self.assertIsNone(act4.decision_)
        self.assertEqual(len(bdd.Tache.query.all()), 1)
        tache = bdd.Tache.query.one()
        self.assertEqual(tache.timestamp, fp_patch.return_value)
        self.assertEqual(tache.commande, "!open 4")
        self.assertEqual(tache.action, act4)
        tache.delete()
        da_patch.assert_not_called()

        # décision, no base_cooldown, no temporel/perma, no charges ==> no diff
        act11 = bdd.Action(base=ba1, joueur=j1, decision_="ach ja",
                           charges=None)
        act11.add()
        await close_action(act11)
        self.assertIsNone(act11.decision_)
        self.assertIsNone(act11.charges)
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # décision, no base_cooldown, no temporel/perma, 4 charges
        act12 = bdd.Action(base=ba1, joueur=j1, decision_="ach ja", charges=4)
        act12.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await close_action(act12)
        self.assertIsNone(act12.decision_)
        self.assertEqual(act12.charges, 3)
        mock_discord.assert_sent(chan, "Il te reste 3 charge")
        mock_discord.assert_not_sent(chan, "pour cette semaine")
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # décision, no base_cooldown, no temporel/perma, 4 charges, refill WE
        ba1b = bdd.BaseAction(slug="ba1b", instant=False, base_cooldown=0,
                              trigger_debut=bdd.ActionTrigger.open_cond,
                              refill="weekends, other")
        ba1b.add()
        act13 = bdd.Action(base=ba1b, joueur=j1, decision_="ach ja", charges=4)
        act13.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await close_action(act13)
        self.assertIsNone(act13.decision_)
        self.assertEqual(act13.charges, 3)
        mock_discord.assert_sent(chan, "Il te reste 3 charge")
        mock_discord.assert_sent(chan, "pour cette semaine")
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # décision, no base_cooldown, no temporel/perma, 1 charge
        act14 = bdd.Action(id=37, base=ba1, joueur=j1, decision_="ach ja",
                           charges=1)
        act14.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await close_action(act14)
        da_patch.assert_called_once_with(act14)     # deleted
        mock_discord.assert_sent(chan, "Il te reste 0 charge")
        mock_discord.assert_not_sent(chan, "pour cette semaine")
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.reset_mock()

        # décision, no base_cooldown, no temporel/perma, 1 charge, refill WE
        act13 = bdd.Action(base=ba1b, joueur=j1, decision_="ach ja", charges=1)
        act13.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await close_action(act13)
        self.assertIsNone(act13.decision_)
        self.assertEqual(act13.charges, 0)      # not deleted
        mock_discord.assert_sent(chan, "Il te reste 0 charge")
        mock_discord.assert_sent(chan, "pour cette semaine")
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # décision, no base_cooldown, no temporel/perma, 1 charge, refill autre
        ba1c = bdd.BaseAction(slug="ba1c", instant=False, base_cooldown=0,
                              trigger_debut=bdd.ActionTrigger.open_cond,
                              refill="brzzz")
        ba1c.add()
        act13 = bdd.Action(base=ba1c, joueur=j1, decision_="ach ja", charges=1)
        act13.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await close_action(act13)
        self.assertIsNone(act13.decision_)
        self.assertEqual(act13.charges, 0)      # not deleted
        mock_discord.assert_sent(chan, "Il te reste 0 charge")
        mock_discord.assert_not_sent(chan, "pour cette semaine")
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.assert_not_called()

        # deleted et plein d'autre trucs
        batro = bdd.BaseAction(slug="batro", instant=False, base_cooldown=2,
                              trigger_debut=bdd.ActionTrigger.perma,
                              refill="brzzz")
        act14 = bdd.Action(base=ba1, joueur=j1, decision_="ach ja", charges=1)
        act14.add()
        with mock_discord.mock_members_and_chans(j1):
            chan = j1.private_chan
            await close_action(act14)
        da_patch.assert_called_once_with(act14)     # deleted
        mock_discord.assert_sent(chan, "Il te reste 0 charge")
        mock_discord.assert_not_sent(chan, "pour cette semaine")
        self.assertEqual(bdd.Tache.query.all(), [])
        da_patch.reset_mock()


    @mock_bdd.patch_db      # Empty database for this method
    @mock.patch("lgrez.features.gestion_actions.delete_action")
    async def test_get_actions(self, da_patch):
        """Unit tests for gestion_actions.get_actions function."""
        # def get_actions(quoi, trigger, heure=None)
        get_actions = gestion_actions.get_actions
        mock_bdd.add_campsroles()
        j1 = bdd.Joueur(discord_id=1, chan_id_=11, nom="Joueur1")
        j1.add()
        perma, temporel = bdd.ActionTrigger.perma, bdd.ActionTrigger.temporel
        bap = bdd.BaseAction(slug="bap", trigger_debut=perma,
                             trigger_fin=perma)
        bat = bdd.BaseAction(slug="bat", trigger_debut=temporel,
                             heure_debut=datetime.time(15, 4),
                             trigger_fin=temporel,
                             heure_fin=datetime.time(22, 15))
        bdd.BaseAction.add(bap, bat)
        act_pn = bdd.Action(id=1, base=bap, joueur=j1, decision_=None)
        act_pr = bdd.Action(id=2, base=bap, joueur=j1, decision_="rien")
        act_pq = bdd.Action(id=3, base=bap, joueur=j1, decision_="qqch")
        act_tn = bdd.Action(id=4, base=bat, joueur=j1, decision_=None)
        act_tr = bdd.Action(id=5, base=bat, joueur=j1, decision_="rien")
        act_tq = bdd.Action(id=6, base=bat, joueur=j1, decision_="qqch")
        bdd.Action.add(act_pn, act_pr, act_pq, act_tn, act_tr, act_tq)

        samples = {
            ("open", bdd.ActionTrigger.mort, None): set(),
            ("close", bdd.ActionTrigger.mort, None): set(),
            ("remind", bdd.ActionTrigger.mort, None): set(),
            ("open", perma, None): {act_pn},
            ("close", perma, None): {act_pr, act_pq},
            ("remind", perma, None): {act_pr},
            ("open", temporel, None): discord.ext.commands.UserInputError,
            ("close", temporel, None): discord.ext.commands.UserInputError,
            ("remind", temporel, None): discord.ext.commands.UserInputError,
            ("open", temporel, datetime.time(15, 4)): {act_tn},
            ("close", temporel, datetime.time(15, 4)): set(),
            ("remind", temporel, datetime.time(15, 4)): set(),
            ("open", temporel, datetime.time(22, 15)): set(),
            ("close", temporel, datetime.time(22, 15)): {act_tr, act_tq},
            ("remind", temporel, datetime.time(22, 15)): {act_tr},
            ("open", temporel, datetime.time(3, 25)): set(),
            ("close", temporel, datetime.time(3, 25)): set(),
            ("remind", temporel, datetime.time(3, 25)): set(),
            ("bzz", perma, None): discord.ext.commands.UserInputError,
        }

        for ((quoi, trigger, heure), result) in samples.items():
            if isinstance(result, type) and issubclass(result, Exception):
                with self.assertRaises(result):
                    if heure:
                        get_actions(quoi, trigger, heure)
                    else:
                        get_actions(quoi, trigger, heure)
            else:
                if heure:
                    actions = get_actions(quoi, trigger, heure)
                else:
                    actions = get_actions(quoi, trigger, heure)
                self.assertEqual(set(actions), result)



import unittest
from unittest import mock

from lgrez import config
from lgrez.bdd import model_jeu
from test import mock_discord, mock_bdd, mock_env



class TestRole(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_jeu.Role methods."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test___repr__(self):
        """Unit tests for Role.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_jeu.Role.__repr__
        slf = mock.Mock(model_jeu.Role, slug="slùgz",
                        prefixe="p' ", nom="n@@mz")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Role 'slùgz' (p' n@@mz)>")


    def test_nom_complet(self):
        """Unit tests for Role.nom_complet property."""
        # @property def nom_complet(self)
        nom_complet = model_jeu.Role.nom_complet
        slf = mock.MagicMock(model_jeu.Role)
        nom = nom_complet.fget(slf)
        self.assertEqual(nom, f"{slf.prefixe}{slf.nom}")
        self.assertIsNone(nom_complet.fset)       # read-only
        self.assertIsNone(nom_complet.fdel)


    def test_default(self):
        """Unit tests for Role.default classmethod."""
        # @classmethod def default(cls)
        default = model_jeu.Role.default
        qry_patch = config.session.query.return_value
        # existing
        cls = mock.MagicMock()
        role = default()
        qry_patch.get.assert_called_once_with(config.default_role_slug)
        qry_patch.reset_mock()
        self.assertEqual(role, qry_patch.get.return_value)
        # existing - other than default
        cls = mock.MagicMock()
        _drs = config.default_role_slug
        config.default_role_slug = "bzbzbz"
        role = default()
        qry_patch.get.assert_called_once_with("bzbzbz")
        qry_patch.reset_mock()
        self.assertEqual(role, qry_patch.get.return_value)
        config.default_role_slug = _drs
        # not existing
        cls = mock.MagicMock()
        qry_patch.get.return_value = None
        with self.assertRaises(ValueError):
            default()
        qry_patch.get.assert_called_once_with(config.default_role_slug)


class TestCamp(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_jeu.Camp methods."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test___repr__(self):
        """Unit tests for Camp.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_jeu.Camp.__repr__
        slf = mock.Mock(model_jeu.Camp, slug="slùgz", nom="n@@mz")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<Camp 'slùgz' (n@@mz)>")


    def test_discord_emoji(self):
        """Unit tests for Camp.discord_emoji property."""
        # @property def discord_emoji(self)
        discord_emoji = model_jeu.Camp.discord_emoji
        config.guild.emojis = [mock.Mock() for i in range(5)]
        for emo, name in zip(config.guild.emojis,
                             ["bz", "op", "glu", "fzoop", "paf"]):
            emo.configure_mock(name=name)
        # pas d'emoji prévu
        slf = mock.MagicMock(model_jeu.Camp, emoji=None)
        with self.assertRaises(ValueError):
            discord_emoji.fget(slf)
        # pas d'emoji sur le serv
        slf = mock.MagicMock(model_jeu.Camp, emoji="bloup")
        with self.assertRaises(ValueError):
            discord_emoji.fget(slf)
        # emoji
        slf = mock.MagicMock(model_jeu.Camp, emoji="fzoop")
        emoji = discord_emoji.fget(slf)
        self.assertEqual(emoji, config.guild.emojis[3])
        # read-only
        self.assertIsNone(discord_emoji.fset)
        self.assertIsNone(discord_emoji.fdel)


    def test_discord_emoji_or_none(self):
        """Unit tests for Camp.discord_emoji_or_none property."""
        # @property def discord_emoji_or_none(self)
        discord_emoji_or_none = model_jeu.Camp.discord_emoji_or_none
        # emoji
        slf = mock.MagicMock(model_jeu.Camp)
        emoji = discord_emoji_or_none.fget(slf)
        self.assertEqual(emoji, slf.discord_emoji)
        # pas d'emoji
        class PropertyMock(mock.MagicMock):
            @property
            def discord_emoji(self):
                raise ValueError
        slf = PropertyMock(model_jeu.Camp)
        emoji = discord_emoji_or_none.fget(slf)
        self.assertIsNone(emoji)
        # read-only
        self.assertIsNone(discord_emoji_or_none.fset)
        self.assertIsNone(discord_emoji_or_none.fdel)


    def test_default(self):
        """Unit tests for Camp.default classmethod."""
        # @classmethod def default(cls)
        default = model_jeu.Camp.default
        qry_patch = config.session.query.return_value
        # existing
        cls = mock.MagicMock()
        camp = default()
        qry_patch.get.assert_called_once_with(config.default_camp_slug)
        qry_patch.reset_mock()
        self.assertEqual(camp, qry_patch.get.return_value)
        # existing - other than default
        cls = mock.MagicMock()
        _drs = config.default_camp_slug
        config.default_camp_slug = "bzbzbz"
        camp = default()
        qry_patch.get.assert_called_once_with("bzbzbz")
        qry_patch.reset_mock()
        self.assertEqual(camp, qry_patch.get.return_value)
        config.default_camp_slug = _drs
        # not existing
        cls = mock.MagicMock()
        qry_patch.get.return_value = None
        with self.assertRaises(ValueError):
            default()
        qry_patch.get.assert_called_once_with(config.default_camp_slug)


class TestBaseAction(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.bdd.model_jeu.BaseAction methods."""

    def test___repr__(self):
        """Unit tests for BaseAction.__repr__ method."""
        # def __repr__(self)
        __repr__ = model_jeu.BaseAction.__repr__
        slf = mock.Mock(model_jeu.BaseAction, slug="slùgz")
        rpr = __repr__(slf)
        self.assertEqual(rpr, "<BaseAction 'slùgz'>")



import unittest

from lgrez import config
from lgrez.blocs import ready_check


class TestConfigGetReady(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.config ready checks."""

    async def test_variables_notready(self):
        """Unit tests for config variables."""

        with self.assertRaises(ready_check.NotReadyError):
            config.guild
        with self.assertRaises(ready_check.NotReadyError):
            config.bot
        with self.assertRaises(ready_check.NotReadyError):
            config.loop
        with self.assertRaises(ready_check.NotReadyError):
            config.engine
        with self.assertRaises(ready_check.NotReadyError):
            config.session

        with self.assertRaises(ready_check.NotReadyError):
            config.Role.mj

        with self.assertRaises(ready_check.NotReadyError):
            config.Channel.logs

        with self.assertRaises(ready_check.NotReadyError):
            config.Emoji.bucher


    async def test___getattr__(self):
        """Unit tests for accessing config variables."""
        __getattr__ = config.__getattr__

        with self.assertRaises(ready_check.NotReadyError):
            __getattr__("guild")

        with self.assertRaises(AttributeError):
            __getattr__("bzzt")



import unittest
from unittest import mock

import discord

from lgrez import config
from lgrez.blocs import one_command
from test import mock_discord



class Test_Bypasser(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.one_command._Bypasser class."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    def test___init__(self):
        """Unit tests for _Bypasser.__init__ method."""
        # def __init__(self, ctx)
        __init__ = one_command._Bypasser.__init__
        # unique case
        slf = mock.Mock(one_command._Bypasser)
        ctx = mock.Mock()
        __init__(slf, ctx)
        self.assertEqual(slf.ctx, ctx)

    def test___enter__(self):
        """Unit tests for _Bypasser.__enter__ method."""
        # def __enter__(self)
        __enter__ = one_command._Bypasser.__enter__
        # unique case
        slf = mock.Mock(one_command._Bypasser, ctx=mock.Mock())
        res = __enter__(slf)
        config.bot.in_command.remove.assert_called_once_with(
            slf.ctx.channel.id)
        self.assertEqual(res, slf)

    def test___exit__(self):
        """Unit tests for _Bypasser.__exit__ method."""
        # def __exit__(self, exc_type, exc, tb)
        __exit__ = one_command._Bypasser.__exit__
        # unique case
        slf = mock.Mock(one_command._Bypasser, ctx=mock.Mock())
        res = __exit__(slf, mock.Mock(), mock.Mock(), mock.Mock())
        config.bot.in_command.append.assert_called_once_with(
            slf.ctx.channel.id)


class TestOneCommandFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.one_command functions."""

    def setUp(self):
        mock_discord.mock_config()

    def tearDown(self):
        mock_discord.unmock_config()

    @mock.patch("lgrez.blocs.one_command._Bypasser")
    def test_bypass(self, bp_class):
        """Unit tests for one_command.bypass function."""
        # def bypass(ctx)
        bypass = one_command.bypass
        # unique case
        slf = mock.Mock()
        ret = bypass(slf)
        bp_class.assert_called_once_with(slf)
        self.assertEqual(ret, bp_class.return_value)


    @mock.patch("asyncio.sleep")
    async def test_not_in_command(self, zzz_patch):
        """Unit tests for one_command.not_in_command function."""
        # async def not_in_command(ctx)
        not_in_command = one_command.not_in_command
        one_command.exempted = ["ah", "oh"]
        config.bot.in_command = [14, 15]

        # exempted
        ctx = mock.Mock(command="ah")
        res = await not_in_command(ctx)
        ctx.send.assert_not_called()
        zzz_patch.assert_not_called()
        self.assertEqual(res, True)

        # channel free
        ctx = mock.Mock(command="eh")
        ctx.channel.id = 13             # not in command
        res = await not_in_command(ctx)
        ctx.send.assert_not_called()
        zzz_patch.assert_not_called()
        self.assertEqual(res, True)

        # no freeing
        ctx = mock.Mock(command="eh", send=mock.AsyncMock())
        ctx.channel.id = 14             # in command
        with self.assertRaises(one_command.AlreadyInCommand):
            await not_in_command(ctx)
        ctx.send.assert_called_once_with(config.stop_keywords[0],
                                         delete_after=0)
        zzz_patch.assert_called_once_with(1)
        self.assertIn(14, config.bot.in_command)
        zzz_patch.reset_mock()

        # freeing!
        ctx = mock.Mock(command="eh", send=mock.AsyncMock())
        ctx.channel.id = 14             # in command
        zzz_patch.side_effect = lambda v: config.bot.in_command.remove(14)
        res = await not_in_command(ctx)
        ctx.send.assert_called_once_with(config.stop_keywords[0],
                                         delete_after=0)
        zzz_patch.assert_called_once_with(1)
        self.assertNotIn(14, config.bot.in_command)
        self.assertEqual(res, True)


    async def test_add_to_in_command(self):
        """Unit tests for one_command.add_to_in_command function."""
        # async def add_to_in_command(ctx)
        add_to_in_command = one_command.add_to_in_command
        one_command.exempted = ["ah", "oh"]
        config.bot.in_command = [14, 15]

        # exempted
        ctx = mock.Mock(command="ah")
        ctx.channel.id = 13             # not in command
        await add_to_in_command(ctx)
        self.assertNotIn(13, config.bot.in_command)     # nothing changed

        # webhook
        ctx = mock.Mock(command="eh")
        ctx.channel.id = 13             # not in command
        await add_to_in_command(ctx)
        self.assertNotIn(13, config.bot.in_command)     # nothing changed

        # other
        ctx = mock.Mock(command="eh")
        ctx.channel.id = 13             # not in command
        ctx.message.webhook_id = None
        await add_to_in_command(ctx)
        self.assertIn(13, config.bot.in_command)        # appended


    @mock.patch("asyncio.sleep")
    async def test_remove_from_in_command(self, zzz_patch):
        """Unit tests for one_command.remove_from_in_command function."""
        # async def remove_from_in_command(ctx)
        remove_from_in_command = one_command.remove_from_in_command
        one_command.exempted = ["ah", "oh"]
        config.bot.in_command = [14, 15]

        # in command but exempted
        ctx = mock.Mock(command="ah")
        ctx.channel.id = 14             # not in command
        await remove_from_in_command(ctx)
        zzz_patch.assert_called_once_with(0.1)
        self.assertIn(14, config.bot.in_command)        # nothing changed
        zzz_patch.reset_mock()

        # not in command
        ctx = mock.Mock(command="eh")
        ctx.channel.id = 13             # not in command
        await remove_from_in_command(ctx)
        zzz_patch.assert_called_once_with(0.1)
        self.assertNotIn(13, config.bot.in_command)     # nothing changed
        zzz_patch.reset_mock()

        # other
        ctx = mock.Mock(command="eh")
        ctx.channel.id = 14             # not in command
        await remove_from_in_command(ctx)
        zzz_patch.assert_called_once_with(0.1)
        self.assertNotIn(14, config.bot.in_command)     # removed


    async def test_do_not_limit(self):
        """Unit tests for one_command.do_not_limit function."""
        # def do_not_limit(command)
        do_not_limit = one_command.do_not_limit
        one_command.exempted = ["ah", "oh"]

        # bad type
        command = mock.Mock()
        with self.assertRaises(TypeError):
            do_not_limit(command)

        # okay
        command = mock.Mock(discord.ext.commands.Command)
        res = do_not_limit(command)
        self.assertIn(command, one_command.exempted)    # apenned
        self.assertEqual(res, command)



import enum
import unittest
from unittest import mock

from lgrez import bdd
from lgrez.blocs import gsheets
from test import mock_env



class TestModif(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.gsheets.Modif class."""

    def test___init__(self):
        """Unit tests for Modif.__init__ method."""
        # def __init__(self, row, column, val)
        __init__ = gsheets.Modif.__init__
        # unique case
        slf = mock.Mock(gsheets.Modif)
        row, column, val = mock.Mock(), mock.Mock(), mock.Mock()
        __init__(slf, row, column, val)
        self.assertEqual(slf.row, row)
        self.assertEqual(slf.column, column)
        self.assertEqual(slf.val, val)

    def test___repr__(self):
        """Unit tests for Modif.__repr__ method."""
        # def __repr__(self)
        __repr__ = gsheets.Modif.__repr__
        # unique case
        slf = mock.Mock(gsheets.Modif, row=12, column=2, val="hohoho")
        res = __repr__(slf)
        self.assertEqual(res, f"<gsheets.Modif: (12, 2) = 'hohoho'>")

    def test___eq__(self):
        """Unit tests for Modif.__eq__ method."""
        # def __eq__(self, other)
        __eq__ = gsheets.Modif.__eq__
        slf = mock.Mock(gsheets.Modif, row=12, column=2, val="hohoho")
        # not same type
        oth0 = "oh"
        self.assertIs(__eq__(slf, oth0), NotImplemented)
        # not equal
        oth1 = mock.Mock(gsheets.Modif, row=11, column=2, val="hohoho")
        oth2 = mock.Mock(gsheets.Modif, row=12, column=1, val="hohoho")
        oth3 = mock.Mock(gsheets.Modif, row=12, column=2, val="hahaha")
        self.assertFalse(__eq__(slf, oth1))
        self.assertFalse(__eq__(slf, oth2))
        self.assertFalse(__eq__(slf, oth3))
        # equal
        oth4 = mock.Mock(gsheets.Modif, row=12, column=2, val="hohoho")
        self.assertTrue(__eq__(slf, oth4))
        self.assertTrue(__eq__(slf, slf))

    def test___hash__(self):
        """Unit tests for Modif.__hash__ method."""
        # def __hash__(self)
        __hash__ = gsheets.Modif.__hash__
        slf = mock.Mock(gsheets.Modif, row=12, column=2, val="hohoho")
        # not same type
        oth0 = "oh"
        self.assertNotEqual(__hash__(slf), hash(oth0))
        # not equal
        oth1 = mock.Mock(gsheets.Modif, row=11, column=2, val="hohoho")
        oth2 = mock.Mock(gsheets.Modif, row=12, column=1, val="hohoho")
        oth3 = mock.Mock(gsheets.Modif, row=12, column=2, val="hahaha")
        self.assertNotEqual(__hash__(slf), __hash__(oth1))
        self.assertNotEqual(__hash__(slf), __hash__(oth2))
        self.assertNotEqual(__hash__(slf), __hash__(oth3))
        # equal
        oth4 = mock.Mock(gsheets.Modif, row=12, column=2, val="hohoho")
        self.assertEqual(__hash__(slf), __hash__(oth4))
        self.assertEqual(__hash__(slf), __hash__(slf))


class TestGsheetsFunctions(unittest.IsolatedAsyncioTestCase):
    """Unit tests for lgrez.blocs.gsheets functions."""

    @mock_env.patch_env(LGREZ_GCP_CREDENTIALS='{"st": 2, "ls": [], "dc": {}}')
    @mock.patch("gspread.authorize")
    @mock.patch("oauth2client.service_account.ServiceAccountCredentials")
    def test_connect(self, sac_patch, auth_patch):
        """Unit tests for gsheets.connect function."""
        # def connect(key)
        connect = gsheets.connect
        # unique case
        key = mock.Mock()
        wb = connect(key)
        sac_patch.from_json_keyfile_dict.assert_called_once_with(
            {"st": 2, "ls": [], "dc": {}},
            ['https://spreadsheets.google.com/feeds']
        )
        auth_patch.assert_called_once_with(
            sac_patch.from_json_keyfile_dict.return_value)
        client = auth_patch.return_value
        client.open_by_key.assert_called_once_with(key)
        self.assertEqual(wb, client.open_by_key.return_value)


    def test_update(self):
        """Unit tests for gsheets.update function."""
        # def update(sheet, *modifs)
        update = gsheets.update

        # 1 modif - str
        sheet = mock.Mock()
        cell = mock.Mock(row=4, col=6)
        sheet.range.return_value = [cell]
        update(sheet, gsheets.Modif(3, 5, "vàl"))
        sheet.range.assert_called_once_with(1, 1, 4, 6)
        self.assertEqual(cell.value, "vàl")
        sheet.update_cells.assert_called_once_with([cell])

        # 1 modif - None
        sheet = mock.Mock()
        cell = mock.Mock(row=4, col=6)
        sheet.range.return_value = [cell]
        update(sheet, gsheets.Modif(3, 5, None))
        sheet.range.assert_called_once_with(1, 1, 4, 6)
        self.assertEqual(cell.value, "")
        sheet.update_cells.assert_called_once_with([cell])

        # 1 modif - int
        sheet = mock.Mock()
        cell = mock.Mock(row=4, col=6)
        sheet.range.return_value = [cell]
        update(sheet, gsheets.Modif(3, 5, 5328))
        sheet.range.assert_called_once_with(1, 1, 4, 6)
        self.assertEqual(cell.value, 5328)
        sheet.update_cells.assert_called_once_with([cell])

        # 1 modif - large int
        sheet = mock.Mock()
        cell = mock.Mock(row=4, col=6)
        sheet.range.return_value = [cell]
        update(sheet, gsheets.Modif(3, 5, 452136988741234))
        sheet.range.assert_called_once_with(1, 1, 4, 6)
        self.assertEqual(cell.value, "452136988741234")
        sheet.update_cells.assert_called_once_with([cell])

        # 1 modif - enum
        class MonEnum(enum.Enum):
            akk = "voui"
        sheet = mock.Mock()
        cell = mock.Mock(row=4, col=6)
        sheet.range.return_value = [cell]
        update(sheet, gsheets.Modif(3, 5, MonEnum.akk))
        sheet.range.assert_called_once_with(1, 1, 4, 6)
        self.assertEqual(cell.value, "akk")
        sheet.update_cells.assert_called_once_with([cell])

        # 1 modif - entrée de BDD
        entree = mock.Mock(bdd.Joueur, primary_key="flebl")
        sheet = mock.Mock()
        cell = mock.Mock(row=4, col=6)
        sheet.range.return_value = [cell]
        update(sheet, gsheets.Modif(3, 5, entree))
        sheet.range.assert_called_once_with(1, 1, 4, 6)
        self.assertEqual(cell.value, "flebl")
        sheet.update_cells.assert_called_once_with([cell])

        # several modifs
        sheet = mock.Mock()
        cells = [mock.Mock(row=4, col=6), mock.Mock(row=4, col=7),
                 mock.Mock(row=1, col=3), mock.Mock(row=8, col=2),]
        sheet.range.return_value = cells
        update(sheet, gsheets.Modif(3, 5, "azz"), gsheets.Modif(3, 6, "bzz"),
               gsheets.Modif(0, 2, "czz"), gsheets.Modif(7, 1, "dzz"))
        sheet.range.assert_called_once_with(1, 1, 8, 7)
        self.assertEqual([cell.value for cell in cells],
                         ["azz", "bzz", "czz", "dzz"])
        sheet.update_cells.assert_called_once_with(cells)


    @mock.patch("gspread.utils.a1_to_rowcol")
    def test_a_to_index(self, gsu_patch):
        """Unit tests for gsheets.a_to_index function."""
        # def a_to_index(column)
        a_to_index = gsheets.a_to_index

        gsu_patch.return_value = (173, 254)
        col = a_to_index("TEST")
        gsu_patch.assert_called_once_with("TEST1")
        self.assertEqual(col, 253)



"""lg-rez / bdd / Briques de connection

Métaclasse et classe de base des tables de données, fonction de connection

"""

import re
import difflib

import sqlalchemy
from sqlalchemy import orm
import unidecode

from lgrez import config
from lgrez.blocs import env


def _remove_accents(text):
    """Renvoie la chaîne non accentuée.

    Mais conserve les caractères spéciaux (emojis...)
    """
    p = re.compile("([À-ʲΆ-ת])")    # Abracadabrax, c'est moche mais ça marche
    return p.sub(lambda m: unidecode.unidecode(m.group()), text)


# ---- Objets de base des classes de données

class TableMeta(sqlalchemy.orm.DeclarativeMeta):
    """Métaclasse des tables de données de LG-Rez.

    Sous-classe de :class:`sqlalchemy.orm.decl_api.DeclarativeMeta`.

    Cette métaclasse :
        - nomme automatiquement la table ``cls.__name__.lower() + "s"``,
        - ajoute une note commune aux docstrings de chaque classe,
        - définit des méthodes et propriétés de classe simplifiant
          l'utilisation des tables.
    """
    def __init__(cls, name, bases, dic, comment=None, **kwargs):
        """Constructs the data class"""
        if name == "TableBase":
            # Ne pas documenter TableBase (pas une vraie table)
            return

        cls.__tablename__ = name.lower() + "s"
        if comment is None:
            comment = cls.__doc__

        dic["registry"] = cls.registry      # un peu de magie noire...
        super().__init__(name, bases, dic, comment=comment, **kwargs)

        cls._attrs = {n: k for n, k in dic.items() if isinstance(k, (
            sqlalchemy.Column,
            sqlalchemy.orm.relationships.RelationshipProperty,
        ))}

        cls.__doc__ += f"""\n    Note:
        Cette classe est une *classe de données* (sous-classe de
        :class:`.TableBase`) représentant la table ``{cls.__tablename__}`` :

        - Les propriétés et méthodes détaillées dans :class:`.TableMeta`
          sont utilisables (sur la classe uniquement) ;
        - Tous les attributs ci-dessous sont du type indiqué pour les
          instances (entrées de BDD), mais de type
          :class:`sqlalchemy.orm.attributes.InstrumentedAttribute`
          pour la classe elle-même.
        """       # On adore la doc vraiment

    @property
    def query(cls):
        """sqlalchemy.orm.query.Query: Raccourci pour
        ``.config.session.query(Table)``.

        Raises:
            :exc:`ready_check.NotReadyError`: session non initialisée
                (:obj:`.config.session` vaut ``None``)
        """
        return config.session.query(cls)

    @property
    def columns(cls):
        """sqlalchemy.sql.base.ImmutableColumnCollection: Raccourci pour
        ``Table.__table__.columns`` (pseudo-dictionnaire nom -> colonne).

        Comportement global de dictionnaire :

            - ``Table.columns["nom"]`` -> colonne associée ;
            - ``Table.columns.keys()`` -> noms des colonnes ;
            - ``Table.columns.values()`` -> objets Column ;
            - ``Table.columns.items()`` -> tuples (nom, colonne) ;

        MAIS itération sur les colonnes (valeurs du dictionnaire) :

            - ``list(Table.columns)`` -> objets Column ;
            - ``for col in Table.columns`` -> objets Column.
        """
        return cls.__table__.columns

    @property
    def attrs(cls):
        """Mapping[:class:`str`, :class:`sqlalchemy.schema.Column` |\
        :class:`sqlalchemy.orm.RelationshipProperty`]: Attributs de
        données publics des instances (dictionnaire nom -> colonne
        / relationship).
        """
        return cls._attrs

    @property
    def primary_col(cls):
        """sqlalchemy.schema.Column: Colonne clé primaire de la table.

        Raises:
            :exc:`ValueError`: Pas de colonne clé primaire pour cette
                table, ou plusieurs
        """
        cols = cls.__table__.primary_key.columns
        if not cols:
            raise ValueError(f"Pas de clé primaire pour {cls.__name__}")
        if len(cols) > 1:
            raise ValueError("Plusieurs colonnes clés primaires pour "
                             f"{cls.__name__} (clé composite)")
        return next(iter(cols))

    def find_nearest(cls, chaine, col=None, sensi=0.25, filtre=None,
                     solo_si_parfait=True, parfaits_only=True,
                     match_first_word=False):
        """Recherche les plus proches résultats d'une chaîne donnée.

        Args:
            chaine (str): motif à rechercher
            col (:class:`sqlalchemy.schema.Column` | :class:`str`):
                colonne selon laquelle rechercher (défaut : colonne
                primaire). Doit être de type textuel.
            sensi (float): score\* minimal pour retenir une entrée
            filtre (~sqlalchemy.sql.expression.BinaryExpression):
                argument de :meth:`query.filter()
                <sqlalchemy.orm.query.Query.filter>`
                (ex. ``Table.colonne == valeur``)
            solo_si_parfait (bool): si ``True`` (défaut), renvoie
                uniquement le premier élément de score\* ``1`` trouvé
                s'il existe (ignore les autres éléments, même si
                ``>= sensi`` ou même ``1``)
            parfaits_only (bool): si ``True`` (défaut), ne renvoie que
                les éléments de score\* ``1`` si on en trouve au moins
                un (ignore les autres éléments, même si ``>= sensi`` ;
                pas d'effet si ``solo_si_parfait`` vaut ``True``)
            match_first_word (bool): si ``True``, teste aussi
                ``chaine`` vis à vis du premier *mot* (caractères
                précédentla première espace) de chaque entrée, et
                conserve ce score si il est supérieur.

        Returns:
            :class:`list`\[\(:class:`TableBase`, :class:`float`\)\]: Les
            entrées correspondant le mieux à ``chaine``, sous forme de
            liste de tuples ``(element, score*)`` triés par score\*
            décroissant (et ce même si un seul résultat).

        Raises:
            ValueError: ``col`` inexistante ou pas de type textuel
            ~ready_check.NotReadyError: session non initialisée
                (:attr:`.lgrez.config.session` vaut ``None``)

        *\*score* = ratio de :class:`difflib.SequenceMatcher`, i.e.
        proportion de caractères communs aux deux chaînes.

        Note:
            Les chaînes sont comparées sans tenir compte de
            l'accentuation ni de la casse.
        """
        if not col:
            col = cls.primary_col
        elif isinstance(col, str):
            try:
                col = cls.columns[col]
            except LookupError:
                raise ValueError(
                    f"{cls.__name__}.find_nearest: Colonne '{col}' invalide"
                ) from None

        if not isinstance(col.type, sqlalchemy.String):
            raise ValueError(f"{cls.__name__}.find_nearest: "
                             f"Colonne {col.key} pas de type textuel")

        query = cls.query
        if filtre is not None:
            query = query.filter(filtre)

        results = query.all()

        SM = difflib.SequenceMatcher()
        # Première chaîne à comparer : cible, en minuscule et sans accents
        slug1 = _remove_accents(chaine).lower()
        SM.set_seq1(slug1)

        scores = []
        for entry in results:
            slug2 = _remove_accents(getattr(entry, col.key)).lower()

            # On compare chaque élément à la cible (en non accentué)
            SM.set_seq2(slug2)
            score = SM.ratio()          # Taux de ressemblance

            if match_first_word:
                # On compare aussi avec le premier mot, si demandé
                first_word = slug2.split(maxsplit=1)[0]
                SM.set_seq2(first_word)
                score_fw = SM.ratio()
                score = max(score, score_fw)

            if score == 1:
                # Cas particulier : élément demandé correspond exactement
                # à un élément existant
                if solo_si_parfait:
                    # Si demandé, on le renvoie direct
                    return [(entry, score)]
                elif parfaits_only:
                    # Si demandé, on ne renverra que les éléments de score 1
                    sensi = 1

            scores.append((entry, score))

        # On ne garde que les résultats >= sensi, dans l'ordre ;
        # si élément parfait trouvé et parfaits_only, sensi = 1 donc
        # ne renvoie que les parfaits
        bests = [(e, score) for (e, score) in scores if score >= sensi]
        return sorted(bests, key=lambda x: x[1], reverse=True)


# Dictionnaire {nom de la base -> table}, automatiquement rempli par
# sqlalchemy.orm.declarative_base
tables = {}


mapper_registry = sqlalchemy.orm.registry(class_registry=tables)

@mapper_registry.as_declarative_base(metaclass=TableMeta)
class TableBase:
    """Classe de base des tables de données.

    (construite par :func:`sqlalchemy.orm.registry.generate_base`)
    """
    @property
    def primary_key(self):
        """Any: Clé primaire de l'instance (``id``, ``slug``...).

        Raccourci pour ``getattr(inst, type(inst).primary_col.key)``.

        Raises:
            :exc:`ValueError`: Pas de colonne clé primaire pour la
                table de cette instance, ou plusieurs.
        """
        key = type(self).primary_col.key
        return getattr(self, key)

    @staticmethod
    def update():
        """Applique les modifications en attente en base (commit).

        Toutes les modifications, y compris des autres instances,
        seront enregistrées.

        Globlament équivalent à::

            config.session.commit()

        """
        config.session.commit()

    def add(self, *other):
        """Enregistre l'instance dans la base de donnée et commit.

        Semble équivalent à :meth:`update` si l'instance est déjà
        présente en base.

        Args:
            \*other: autres instances à enregistrer dans le même commit,
                éventuellement. Utilisation recommandée :
                ``<Table>.add(*items)``.

        Examples:
            - ``item.add()``
            - ``<Table>.add(*items)``

        Globlament équivalent à::

            config.session.add(self)
            config.session.add_all(other)

            config.session.commit()
        """
        config.session.add(self)
        if other:
            config.session.add_all(other)

        self.update()

    def delete(self, *other):
        """Supprime l'instance de la base de données et commit.

        Args:
            \*other: autres instances à supprimer dans le même commit,
                éventuellement.

        Examples:
            - ``item.delete()``
            - ``<Table>.delete(*items)``

        Raises:
            sqlalchemy.exc.SAWarning: l'instance a déjà été supprimmée
                (Warning, pas exception : ne bloque pas l'exécution).

        Globlament équivalent à::

            config.session.delete(self)
            for item in others:
                config.session.delete(item)

            config.session.commit()
        """
        config.session.delete(self)
        for item in other:
            config.session.delete(item)

        self.update()


# ---- Autodoc objects

def autodoc_Column(*args, doc="", comment=None, **kwargs):
    """Returns Python-side and SQL-side well documented Column.

    Use exactly as :class:`sqlalchemy.Column <sqlalchemy.schema.Column>`.

    Args:
        \*args, \*\*kwargs: passed to :class:`sqlalchemy.schema.Column`.
        doc (str): column description, enhanced with column infos
            (Python and SQL types, if nullable, primary...) and
            passed to ``Column.doc``.
        comment (str): passed to ``Column.comment``; defaults to ``doc``
            (not enhanced). Set it to ``''`` to disable comment creation.

    Returns:
        :class:`sqlalchemy.schema.Column`
    """
    if comment is None:
        comment = doc
    col = sqlalchemy.Column(*args, **kwargs, comment=comment)
    sa_type = (f":class:`{col.type!r} "
               f"<sqlalchemy.types.{type(col.type).__name__}>`")
    py_type = col.type.python_type
    if py_type.__module__ in ("builtins", "lgrez.bdd.enums"):
        py_type_str = py_type.__name__
    elif py_type.__module__.startswith("lgrez.bdd.model_"):
        py_type_str = f"lgrez.bdd.{py_type.__name__}"
    else:
        py_type_str = f"{py_type.__module__}.{py_type.__name__}"
    or_none = " | ``None``" if col.nullable else ""
    primary = " (clé primaire)" if col.primary_key else ""
    autoinc = (" (auto-incrémental)"
               if (col.autoincrement and primary
                   and isinstance(col.type, sqlalchemy.Integer))
               else "")
    nullable = "" if (col.nullable or autoinc) else " (NOT NULL)"
    default = f" (défaut ``{col.default.arg!r}``)" if col.default else ""
    col.doc = (f"{doc}{primary}{autoinc}{nullable}{default}\n\n"
               f"Type SQLAlchemy : {sa_type} / Type SQL : ``{col.type}``\n\n"
               f":type: :class:`{py_type_str}`{or_none}")
    return col


def autodoc_ManyToOne(tablename, *args, doc="", nullable=False, **kwargs):
    """Returns Python-side well documented many-to-one relationship.

    Represents a relationship where each element in this table is
    linked to **one element** in ``tablename``, itself back-refering
    to **several elements** in this table.

    Example: ``Book.editor`` (each book has an editor, an editor
    publishes several books)

    Use exactly as :func:`sqlalchemy.orm.relationship`, plus the
    keyword argument :attr:`nullable` (see below).

    Args:
        \*args, \*\*kwargs: passed to :class:`sqlalchemy.orm.relationship`.
        doc (str): relationship description, enhanced with class name
            and relationship type.
        nullable (bool): indicates whether the relationship can be
            omitted (impacts docs only, default ``False``).

    Returns:
        :class:`sqlalchemy.orm.RelationshipProperty`
    """
    first, sep, rest = doc.partition("\n")
    or_none = " | ``None``" if nullable else ""
    doc = (f":class:`~.bdd.{tablename}`{or_none}: {first} "
           "*(many-to-one relationship)*")
    if not nullable:
        doc += " (NOT NULL)"
    if rest:
        doc += sep + rest

    return sqlalchemy.orm.relationship(tablename, *args, doc=doc, **kwargs)


def autodoc_OneToMany(tablename, *args, doc="", **kwargs):
    """Returns Python-side well documented one-to-many relationship.

    Represents a relationship where each element in this table is
    linked to **several elements** in ``tablename``, itself back-
    refering to **one element** in this table.

    Example: ``Editor.books`` (an editor publishes several books,
    each book has an editor)

    Use exactly as :func:`sqlalchemy.orm.relationship`.

    Args:
        \*args, \*\*kwargs: passed to :class:`sqlalchemy.orm.relationship`.
        doc (str): relationship description, enhanced with class name
            and relationship type.

    Returns:
        :class:`sqlalchemy.orm.RelationshipProperty`
    """
    first, sep, rest = doc.partition("\n")
    doc = f"Sequence[~bdd.{tablename}]: {first} *(one-to-many relationship)*"
    if rest:
        doc += sep + rest

    return sqlalchemy.orm.relationship(tablename, *args, doc=doc, **kwargs)


def autodoc_DynamicOneToMany(tablename, *args, doc="", **kwargs):
    """Returns Python-side well documented dynamic one-to-many relationship.

    Represents a relationship where each element in this table is
    linked to **a lot of elements** in ``tablename``, itself back-
    refering to **one element** in this table.

    The difference with :func:`~.autodoc_OneToMany` is that items are NOT
    accessible directly through class attribute, which returns a query
    that but must be fetched first:

    Examples: ``Editor.books.all()``, ``Editor.books.filter_by(...).one()``

    Use exactly as :func:`sqlalchemy.orm.dynamic_loader`.

    Args:
        \*args, \*\*kwargs: passed to :class:`sqlalchemy.orm.dynamic_loader`.
        doc (str): relationship description, enhanced with class name
            and relationship type.

    Returns:
        :class:`sqlalchemy.orm.RelationshipProperty`
    """
    first, sep, rest = doc.partition("\n")
    doc = (f"~sqlalchemy.orm.query.Query[~bdd.{tablename}]: {first} "
           "*(dynamic one-to-many relationship)*")
    if rest:
        doc += sep + rest

    return sqlalchemy.orm.dynamic_loader(tablename, *args, doc=doc, **kwargs)


def autodoc_ManyToMany(tablename, *args, doc="", **kwargs):
    """Returns Python-side well documented many-to-many relationship.

    Represents a relationship where each element in this table is
    linked to **several elements** in ``tablename``, itself back-
    refering to **several elements** in this table.

    Example: ``Book.authors`` (each book has several authors, an
    author writes several books)

    Use exactly as :func:`sqlalchemy.orm.relationship`.

    Args:
        \*args, \*\*kwargs: passed to :class:`sqlalchemy.orm.relationship`.
        doc (str): relationship description, enhanced with class name
            and relationship type.

    Returns:
        :class:`sqlalchemy.orm.RelationshipProperty`
    """
    first, sep, rest = doc.partition("\n")
    doc = f"Sequence[~bdd.{tablename}]: {first} *(many-to-many relationship)*"
    if rest:
        doc += sep + rest

    return sqlalchemy.orm.relationship(tablename, *args, doc=doc, **kwargs)


# ---- Connection function

def connect():
    """Se connecte à la base de données et prépare les objets connectés.

    - Utilise la variable d'environment ``LGREZ_DATABASE_URI``
    - Crée les tables si nécessaire
    - Prépare :obj:`.config.engine` et :obj:`.config.session`
    """
    LGREZ_DATABASE_URI = env.load("LGREZ_DATABASE_URI")
    # Moteur SQL : connexion avec le serveur
    config.engine = sqlalchemy.create_engine(LGREZ_DATABASE_URI,
                                             pool_pre_ping=True)

    # Création des tables si elles n'existent pas déjà
    TableBase.metadata.create_all(config.engine)

    # Ouverture de la session
    Session = sqlalchemy.orm.sessionmaker(bind=config.engine)
    config.session = Session()
